import logging
import math

import numpy as np
import pandas as pd

import wind_up
from wind_up.constants import (
    CONFIG_DIR,
    PROJECTROOT_DIR,
    RANDOM_SEED,
    REANALYSIS_WD_COL,
    REANALYSIS_WS_COL,
    WINDFARM_YAWDIR_COL,
    DataColumns,
)
from wind_up.detrend import apply_wsratio_v_wd_scen, calc_wsratio_v_wd_scen, check_applied_detrend
from wind_up.interface import AssessmentInputs, add_toggle_signals
from wind_up.long_term import calc_turbine_lt_dfs_raw_filt
from wind_up.math_funcs import circ_diff
from wind_up.models import PlotConfig, Turbine, WindUpConfig
from wind_up.northing import (
    check_wtg_northing,
)
from wind_up.plots.data_coverage_plots import plot_detrend_data_cov, plot_pre_post_data_cov
from wind_up.plots.detrend_plots import plot_apply_wsratio_v_wd_scen
from wind_up.plots.scada_funcs_plots import compare_ops_curves_pre_post, print_filter_stats
from wind_up.plots.yaw_direction_plots import plot_yaw_direction_pre_post
from wind_up.pp_analysis import pre_post_pp_analysis_with_reversal_and_bootstrapping
from wind_up.result_manager import result_manager
from wind_up.waking_state import (
    add_waking_scen,
    get_distance_and_bearing,
    get_iec_upwind_turbines,
    lat_long_is_valid,
)
from wind_up.windspeed_drift import check_windspeed_drift

logger = logging.getLogger(__name__)


def get_config_objects(
    config_yaml_file_name: str,
    *,
    show_plots: bool,
    save_plots: bool,
) -> tuple[WindUpConfig, PlotConfig]:
    cfg = WindUpConfig.from_yaml(CONFIG_DIR / config_yaml_file_name)
    plot_cfg = PlotConfig(show_plots=show_plots, save_plots=save_plots, plots_dir=cfg.out_dir / "plots")
    return cfg, plot_cfg


def add_fake_power_data(
    ref_df: pd.DataFrame,
    *,
    ref_pw_col: str,
    ref_ws_col: str,
    scada_pc: pd.DataFrame,
) -> pd.DataFrame:
    ref_df[ref_pw_col] = np.interp(ref_df[ref_ws_col], scada_pc["WindSpeedMean"], scada_pc["pw_clipped"])
    return ref_df


def get_ref_lat_long(ref_name: str, cfg: WindUpConfig) -> tuple[float, float]:
    try:
        ref_wtg = next(x for x in cfg.asset.wtgs if x.name == ref_name)
        ref_lat = ref_wtg.latitude
        ref_long = ref_wtg.longitude
    except StopIteration as exc:
        if ref_name == "reanalysis":
            ref_lat = np.nan
            ref_long = np.nan
        elif ref_name in [x.name for x in cfg.asset.masts_and_lidars]:
            ref_obj = next(x for x in cfg.asset.masts_and_lidars if x.name == ref_name)
            ref_lat = ref_obj.latitude
            ref_long = ref_obj.longitude
        else:
            msg = "ref_name must be a wtg name or 'reanalysis' or exist in cfg.asset.masts_and_lidars"
            raise ValueError(msg) from exc
    return ref_lat, ref_long


def filter_ref_df_for_wake_free(
    ref_df: pd.DataFrame,
    *,
    ref_name: str,
    ref_wd_col: str,
    cfg: WindUpConfig,
    test_wtg: Turbine,
) -> pd.DataFrame:
    ref_df = ref_df.copy()
    ref_df["rounded_wd"] = ref_df[ref_wd_col].round(0) % 360
    if cfg.require_test_wake_free:
        rows_before = len(ref_df)
        test_latlongs = test_wtg.get_latlongs()
        if len(test_latlongs) == 1 and lat_long_is_valid(test_latlongs[0][0], test_latlongs[0][1]):
            test_wds_to_keep = []
            for wd in ref_df["rounded_wd"].unique():
                test_upwind_wtgs = get_iec_upwind_turbines(
                    latlongs=test_latlongs,
                    wind_direction=wd,
                    cfg=cfg,
                    object_name=test_wtg.name,
                )
                if len(test_upwind_wtgs) == 0:
                    test_wds_to_keep.append(wd)
            ref_df = ref_df[ref_df["rounded_wd"].isin(test_wds_to_keep)]
            if len(test_wds_to_keep) > 0:
                logger.info(
                    f"{test_wtg.name} wake free directions min={min(test_wds_to_keep)} max={max(test_wds_to_keep)}"
                )
            else:
                logger.info(f"{test_wtg.name} has no wake free directions with data")
        rows_after = len(ref_df)
        logger.info(
            f"removed {rows_before - rows_after} [{100 * (rows_before - rows_after) / rows_before:.1f}%] "
            f"rows from ref_df using require_test_wake_free",
        )
    if cfg.require_ref_wake_free:
        rows_before = len(ref_df)
        ref_lat, ref_long = get_ref_lat_long(ref_name, cfg)
        if lat_long_is_valid(ref_lat, ref_long):
            ref_wds_to_keep = []
            for wd in ref_df["rounded_wd"].unique():
                ref_upwind_wtgs = get_iec_upwind_turbines(
                    latlongs=[(ref_lat, ref_long)],
                    wind_direction=wd,
                    cfg=cfg,
                    object_name=ref_name,
                )
                if len(ref_upwind_wtgs) == 0:
                    ref_wds_to_keep.append(wd)
            ref_df = ref_df[ref_df["rounded_wd"].isin(ref_wds_to_keep)]
            if len(ref_wds_to_keep) > 0:
                logger.info(
                    f"{ref_name} wake free directions with data min={min(ref_wds_to_keep)} max={max(ref_wds_to_keep)}",
                )
            else:
                logger.info(f"{ref_name} has no wake free directions with data")
        rows_after = len(ref_df)
        logger.info(
            f"removed {rows_before - rows_after} [{100 * (rows_before - rows_after) / rows_before:.1f}%] "
            f"rows from ref_df using require_ref_wake_free",
        )
    return ref_df.drop(columns=["rounded_wd"])


def filter_ref_df_for_wd_and_hod(ref_df: pd.DataFrame, ref_wd_col: str, cfg: WindUpConfig) -> pd.DataFrame:
    ref_df = ref_df.copy()
    if cfg.ref_wd_filter is not None:
        rows_before = len(ref_df)
        if cfg.ref_wd_filter[0] < cfg.ref_wd_filter[1]:
            ref_df = ref_df[(ref_df[ref_wd_col] >= cfg.ref_wd_filter[0]) & (ref_df[ref_wd_col] <= cfg.ref_wd_filter[1])]
        else:
            ref_df = ref_df[(ref_df[ref_wd_col] >= cfg.ref_wd_filter[0]) | (ref_df[ref_wd_col] <= cfg.ref_wd_filter[1])]
        rows_after = len(ref_df)
        logger.info(
            f"removed {rows_before - rows_after} [{100 * (rows_before - rows_after) / rows_before:.1f}%] "
            f"rows from ref_df using ref_wd_filter",
        )

    if cfg.ref_hod_filter is not None:
        rows_before = len(ref_df)
        if cfg.ref_hod_filter[0] < cfg.ref_hod_filter[1]:
            ref_df = ref_df[(ref_df.index.hour >= cfg.ref_hod_filter[0]) & (ref_df.index.hour <= cfg.ref_hod_filter[1])]
        else:
            ref_df = ref_df[(ref_df.index.hour >= cfg.ref_hod_filter[0]) | (ref_df.index.hour <= cfg.ref_hod_filter[1])]
        rows_after = len(ref_df)
        logger.info(
            f"removed {rows_before - rows_after} [{100 * (rows_before - rows_after) / rows_before:.1f}%] "
            f"rows from ref_df using ref_hod_filter",
        )
    return ref_df


def get_ref_df(
    *,
    ref_name: str,
    wf_df: pd.DataFrame,
    ref_pw_col: str,
    ref_ws_col: str,
    ref_wd_col: str,
    scada_pc: pd.DataFrame,
    cfg: WindUpConfig,
    test_wtg: Turbine,
    toggle_df: pd.DataFrame | None = None,
    keep_only_toggle_off: bool = True,
) -> pd.DataFrame:
    if ref_name in [x.name for x in cfg.asset.wtgs]:
        ref_df = wf_df.loc[ref_name].copy()
        if cfg.toggle is not None:
            ref_df = add_toggle_signals(ref_df, wtg_name=ref_name, cfg=cfg, toggle_df=toggle_df)
            ref_df = ref_df.rename(columns={"toggle_off": "ref_toggle_off", "toggle_on": "ref_toggle_on"})
            if keep_only_toggle_off:
                ref_df = ref_df[ref_df["ref_toggle_off"].fillna(value=False)]
                if ref_df["ref_toggle_on"].any():
                    msg = "keep_only_toggle_off = True but ref_df['ref_toggle_on'] has True values"
                    raise RuntimeError(msg)
    else:
        if ref_name == "reanalysis":
            ref_df = wf_df.loc[cfg.test_wtgs[0].name, [REANALYSIS_WS_COL, REANALYSIS_WD_COL, WINDFARM_YAWDIR_COL]]
            original_ws_col = REANALYSIS_WS_COL
            original_wd_col = REANALYSIS_WD_COL
        elif ref_name in [x.name for x in cfg.asset.masts_and_lidars]:
            ref_obj = next(x for x in cfg.asset.masts_and_lidars if x.name == ref_name)
            ref_df = pd.read_parquet(
                PROJECTROOT_DIR / "input_data" / "masts_and_lidars" / cfg.asset.name / f"{ref_obj.data_file_name}",
            )
            northing_df = wf_df.loc[cfg.test_wtgs[0].name, [REANALYSIS_WS_COL, REANALYSIS_WD_COL, WINDFARM_YAWDIR_COL]]
            ref_df = ref_df.merge(northing_df, how="left", left_index=True, right_index=True)
            original_ws_col = ref_obj.wind_speed_column
            original_wd_col = ref_obj.wind_direction_column
        else:
            msg = "ref_name must be a wtg name or 'reanalysis' or exist in cfg.asset.masts_and_lidars"
            raise ValueError(msg)
        ref_df[ref_ws_col] = ref_df[original_ws_col]
        ref_df["WindSpeedMean"] = ref_df[original_ws_col]  # needed for northing analysis
        ref_df["ws_est_from_power_only"] = ref_df[original_ws_col]  # needed for reversal analysis
        ref_df[ref_wd_col] = ref_df[original_wd_col]
        ref_df = add_fake_power_data(ref_df, ref_pw_col=ref_pw_col, ref_ws_col=ref_ws_col, scada_pc=scada_pc)

    ref_df = filter_ref_df_for_wd_and_hod(ref_df, ref_wd_col=ref_wd_col, cfg=cfg)

    if cfg.require_test_wake_free or cfg.require_ref_wake_free:
        ref_df = filter_ref_df_for_wake_free(
            ref_df,
            ref_name=ref_name,
            ref_wd_col=ref_wd_col,
            cfg=cfg,
            test_wtg=test_wtg,
        )
    return ref_df


def make_extended_time_index(
    original_index: pd.DatetimeIndex,
    *,
    timebase: pd.Timedelta,
    max_timedelta_seconds: int,
) -> pd.DatetimeIndex:
    if not isinstance(original_index, pd.DatetimeIndex):
        msg = f"original_index must be a pd.DatetimeIndex, not {type(original_index)}"
        raise TypeError(msg)
    extended_index = original_index
    timedelta_multiple = -math.floor(max_timedelta_seconds / timebase.total_seconds())
    max_timedelta_multiple = math.floor(max_timedelta_seconds / timebase.total_seconds())
    while timedelta_multiple <= max_timedelta_multiple:
        shifted_index = original_index + (timebase * timedelta_multiple)
        extended_index = pd.DatetimeIndex(pd.concat([pd.Series(extended_index), pd.Series(shifted_index)]))
        timedelta_multiple += 1
    return extended_index.sort_values().drop_duplicates()


def toggle_pairing_filter(
    *,
    pre_df: pd.DataFrame,
    post_df: pd.DataFrame,
    pairing_filter_method: str,
    pairing_filter_timedelta_seconds: int,
    detrend_ws_col: str,
    test_pw_col: str,
    ref_wd_col: str,
    timebase_s: int,
) -> tuple[pd.DataFrame, pd.DataFrame]:
    required_cols = [detrend_ws_col, test_pw_col, ref_wd_col]
    len_pre_before = len(pre_df.dropna(subset=required_cols))
    len_post_before = len(post_df.dropna(subset=required_cols))
    if pairing_filter_method == "none":
        filt_pre_df = pre_df
        filt_post_df = post_df
    elif pairing_filter_method == "one_to_one":
        msg = "pairing_filter_method one_to_one not implemented"
        raise NotImplementedError(msg)
    elif pairing_filter_method == "any_within_timedelta":
        valid_pre_df = pre_df.dropna(subset=required_cols)
        valid_post_df = post_df.dropna(subset=required_cols)
        filt_pre_df = valid_pre_df[
            valid_pre_df.index.isin(
                make_extended_time_index(
                    valid_post_df.index,
                    timebase=pd.Timedelta(seconds=timebase_s),
                    max_timedelta_seconds=pairing_filter_timedelta_seconds,
                ),
            )
        ]
        filt_post_df = valid_post_df[
            valid_post_df.index.isin(
                make_extended_time_index(
                    valid_pre_df.index,
                    timebase=pd.Timedelta(seconds=timebase_s),
                    max_timedelta_seconds=pairing_filter_timedelta_seconds,
                ),
            )
        ]
    else:
        msg = f"pairing_filter_method {pairing_filter_method} not recognised"
        raise ValueError(msg)
    len_pre_after = len(filt_pre_df)
    len_post_after = len(filt_post_df)
    logger.info(
        f"removed {len_pre_before - len_pre_after} [{100 * (len_pre_before - len_pre_after) / len_pre_before:.1f}%] "
        f"rows from pre_df using {pairing_filter_method} pairing filter",
    )
    logger.info(
        f"removed {len_post_before - len_post_after} "
        f"[{100 * (len_post_before - len_post_after) / len_post_before:.1f}%] "
        f"rows from post_df using {pairing_filter_method} pairing filter",
    )
    return filt_pre_df, filt_post_df


def yaw_error_results(pre_df: pd.DataFrame, post_df: pd.DataFrame, required_pp_cols: list[str]) -> dict:
    results = {}
    if "test_yaw_error_mean" in pre_df.columns:
        results["test_yaw_error_pre"] = pre_df.dropna(subset=required_pp_cols)["test_yaw_error_mean"].mean()
        results["test_yaw_error_post"] = post_df.dropna(subset=required_pp_cols)["test_yaw_error_mean"].mean()
    if "ref_yaw_error_mean" in pre_df.columns:
        results["ref_yaw_error_pre"] = pre_df.dropna(subset=required_pp_cols)["ref_yaw_error_mean"].mean()
        results["ref_yaw_error_post"] = post_df.dropna(subset=required_pp_cols)["ref_yaw_error_mean"].mean()
    return results


def yaw_offset_results(
    pre_df: pd.DataFrame, post_df: pd.DataFrame, required_pp_cols: list[str], ref_wd_col: str, test_wd_col: str
) -> dict:
    results = {}

    pre_yaw_offset = pd.Series(
        circ_diff(
            pre_df.dropna(subset=required_pp_cols)[ref_wd_col], pre_df.dropna(subset=required_pp_cols)[test_wd_col]
        )
    )
    post_yaw_offset = pd.Series(
        circ_diff(
            post_df.dropna(subset=required_pp_cols)[ref_wd_col], post_df.dropna(subset=required_pp_cols)[test_wd_col]
        )
    )

    if len(pre_yaw_offset) > 0 and len(post_yaw_offset) > 0:
        results["mean_test_yaw_offset_pre"] = pre_yaw_offset.mean()
        results["mean_test_yaw_offset_post"] = post_yaw_offset.mean()

    yaw_offset_ul = 1e-3
    if "test_yaw_offset_command" in pre_df.columns:
        result_name = "mean_test_yaw_offset_command_pre"
        results[result_name] = pre_df.dropna(subset=required_pp_cols)["test_yaw_offset_command"].mean()
        if results[result_name] > yaw_offset_ul:
            result_manager.warning(f"{result_name} > 0: " f"({results[result_name]})")

        results["mean_test_yaw_offset_command_post"] = post_df.dropna(subset=required_pp_cols)[
            "test_yaw_offset_command"
        ].mean()
    if "ref_yaw_offset_command" in pre_df.columns:
        result_name = "mean_ref_yaw_offset_command_pre"
        results[result_name] = pre_df.dropna(subset=required_pp_cols)["ref_yaw_offset_command"].mean()
        if results[result_name] > yaw_offset_ul:
            result_manager.warning(f"{result_name} > 0 for: " f"({results[result_name]})")

        result_name = "mean_ref_yaw_offset_command_pre"
        results[result_name] = post_df.dropna(subset=required_pp_cols)["ref_yaw_offset_command"].mean()
        if results[result_name] > yaw_offset_ul:
            result_manager.warning(f"{result_name} > 0 for: " f"({results[result_name]})")
    return results


def check_for_ops_curve_shift(
    pre_df: pd.DataFrame,
    post_df: pd.DataFrame,
    *,
    wtg_name: str,
    ws_col: str,
    pw_col: str,
    rpm_col: str,
    pt_col: str,
) -> dict[str, float]:
    results_dict = {
        "powercurve_shift": np.nan,
        "rpm_shift": np.nan,
        "pitch_shift": np.nan,
    }
    # check if all required columns are present
    required_cols = [ws_col, pw_col, pt_col, rpm_col]
    for req_col in required_cols:
        if req_col not in pre_df.columns:
            msg = f"check_for_ops_curve_shift {wtg_name} pre_df missing required column {req_col}"
            result_manager.warning(msg)
            return results_dict
        if req_col not in post_df.columns:
            msg = f"check_for_ops_curve_shift {wtg_name} post_df missing required column {req_col}"
            result_manager.warning(msg)
            return results_dict
    pre_dropna_df = pre_df.dropna(subset=[ws_col, pw_col, pt_col, rpm_col]).copy()
    post_dropna_df = post_df.dropna(subset=[ws_col, pw_col, pt_col, rpm_col]).copy()

    for descr, x_var, y_var, x_bin_width, warn_thresh in [
        ("powercurve_shift", ws_col, pw_col, 1, 0.01),
        ("rpm_shift", pw_col, rpm_col, 0, 0.005),
        ("pitch_shift", ws_col, pt_col, 1, 0.1),
    ]:
        bins = np.arange(0, pre_dropna_df[x_var].max() + x_bin_width, x_bin_width) if x_bin_width > 0 else 10
        mean_curve = pre_dropna_df.groupby(pd.cut(pre_dropna_df[x_var], bins=bins, retbins=False), observed=True).agg(
            x_mean=pd.NamedAgg(column=x_var, aggfunc="mean"),
            y_mean=pd.NamedAgg(column=y_var, aggfunc="mean"),
        )
        post_dropna_df["expected_y"] = np.interp(post_dropna_df[x_var], mean_curve["x_mean"], mean_curve["y_mean"])
        mean_df = post_dropna_df.mean()
        if y_var == pt_col:
            results_dict[descr] = mean_df[y_var] - mean_df["expected_y"]
        else:
            results_dict[descr] = (mean_df[y_var] / mean_df["expected_y"] - 1).clip(-1, 1)
        if abs(results_dict[descr]) > warn_thresh:
            result_manager.warning(
                f"{wtg_name} check_for_ops_curve_shift abs {descr} > {warn_thresh}: {results_dict[descr]}"
            )

    return results_dict


def calc_test_ref_results(
    *,
    test_df: pd.DataFrame,
    pre_df: pd.DataFrame,
    post_df: pd.DataFrame,
    long_term_df: pd.DataFrame | None,
    wf_df: pd.DataFrame,
    test_wtg: Turbine,
    ref_name: str,
    test_pw_col: str,
    test_ws_col: str,
    scada_pc: pd.DataFrame,
    cfg: WindUpConfig,
    plot_cfg: PlotConfig,
    random_seed: int,
    toggle_df: pd.DataFrame | None = None,
) -> dict:
    test_name = test_wtg.name
    ref_pw_col = "pw_clipped"
    if test_name == ref_name:
        ref_ws_col = "WindSpeedMean"
        test_ws_col = "test_WindSpeedMean"
    else:
        ref_ws_col = "ws_est_from_power_only" if cfg.ignore_turbine_anemometer_data else "ws_est_blend"
    ref_info = {
        "ref": ref_name,
    }
    ref_wd_col = "YawAngleMean"
    keep_only_toggle_off = False
    if cfg.toggle is not None:
        # if ref is test then keep all data, not just toggle off
        # if there is not a toggle file per turbine then keep all data
        keep_only_toggle_off = cfg.toggle.toggle_file_per_turbine and (ref_name != test_name)
    ref_df = get_ref_df(
        ref_name=ref_name,
        wf_df=wf_df,
        ref_pw_col=ref_pw_col,
        ref_ws_col=ref_ws_col,
        ref_wd_col=ref_wd_col,
        scada_pc=scada_pc,
        cfg=cfg,
        test_wtg=test_wtg,
        toggle_df=toggle_df,
        keep_only_toggle_off=keep_only_toggle_off,
    )
    if len(ref_df) == 0:
        result_manager.warning(f"ref_df is empty for {ref_name}")
        return ref_info
    ref_max_northing_error_v_reanalysis = check_wtg_northing(
        ref_df,
        wtg_name=ref_name,
        north_ref_wd_col=REANALYSIS_WD_COL,
        timebase_s=cfg.timebase_s,
        plot_cfg=plot_cfg,
    )
    ref_max_northing_error_v_wf = check_wtg_northing(
        ref_df,
        wtg_name=ref_name,
        north_ref_wd_col=WINDFARM_YAWDIR_COL,
        timebase_s=cfg.timebase_s,
        plot_cfg=plot_cfg,
    )

    ref_pw_col = "ref_" + ref_pw_col
    ref_ws_col = "ref_" + ref_ws_col
    ref_wd_col = "ref_" + ref_wd_col
    ref_df.columns = ["ref_" + x for x in ref_df.columns]

    test_wtg = next(x for x in cfg.asset.wtgs if x.name == test_name)
    test_lat = test_wtg.latitude
    test_long = test_wtg.longitude
    ref_lat, ref_long = get_ref_lat_long(ref_name, cfg)

    distance_m, bearing_deg = get_distance_and_bearing(
        lat1=test_lat,
        long1=test_long,
        lat2=ref_lat,
        long2=ref_long,
    )

    ref_max_ws_drift, ref_max_ws_drift_pp_period = check_windspeed_drift(
        wtg_df=ref_df,
        wtg_name=ref_name,
        ws_col=ref_ws_col,
        reanalysis_ws_col="ref_" + REANALYSIS_WS_COL,
        cfg=cfg,
        plot_cfg=plot_cfg,
    )

    detrend_df = test_df.merge(ref_df, how="left", left_index=True, right_index=True)
    detrend_df = detrend_df[cfg.detrend_first_dt_utc_start : cfg.detrend_last_dt_utc_start]  # type: ignore[misc]
    if "test_toggle_on" in detrend_df.columns:
        # find the first time where "test_toggle_on" is true
        first_toggle_on = detrend_df[detrend_df["test_toggle_on"].fillna(value=False)].index.min()
        rows_before = len(detrend_df)
        detrend_df = detrend_df[
            (detrend_df.index < first_toggle_on) | (detrend_df["test_toggle_off"].fillna(value=False))
        ]
        rows_after = len(detrend_df)
        logger.info(
            f"removed {rows_before - rows_after} [{100 * (rows_before - rows_after) / rows_before:.1f}%] "
            f"rows from detrend_df where test_toggle_off was not True after the first toggle on time",
        )
    detrend_df = add_waking_scen(
        test_name=test_name,
        ref_name=ref_name,
        test_ref_df=detrend_df,
        cfg=cfg,
        wf_df=wf_df,
        ref_wd_col=ref_wd_col,
        ref_lat=ref_lat,
        ref_long=ref_long,
    )

    plot_detrend_data_cov(
        cfg=cfg,
        test_name=test_name,
        ref_name=ref_name,
        test_df=test_df,
        test_ws_col=test_ws_col,
        ref_df=ref_df,
        ref_ws_col=ref_ws_col,
        ref_wd_col=ref_wd_col,
        detrend_df=detrend_df,
        plot_cfg=plot_cfg,
    )

    wsratio_v_dir_scen = calc_wsratio_v_wd_scen(
        test_name=test_name,
        ref_name=ref_name,
        ref_lat=ref_lat,
        ref_long=ref_long,
        detrend_df=detrend_df,
        test_ws_col=test_ws_col,
        ref_ws_col=ref_ws_col,
        ref_wd_col=ref_wd_col,
        cfg=cfg,
        plot_cfg=plot_cfg,
    )

    pre_df = pre_df.merge(ref_df, how="left", left_index=True, right_index=True)
    post_df = post_df.merge(ref_df, how="left", left_index=True, right_index=True)

    ref_ops_curve_shift_dict = check_for_ops_curve_shift(
        pre_df,
        post_df,
        wtg_name=ref_name,
        ws_col=ref_ws_col,
        pw_col=ref_pw_col,
        rpm_col=f"ref_{DataColumns.gen_rpm_mean}",
        pt_col=f"ref_{DataColumns.pitch_angle_mean}",
    )

    pre_df = add_waking_scen(
        test_ref_df=pre_df,
        test_name=test_name,
        ref_name=ref_name,
        cfg=cfg,
        wf_df=wf_df,
        ref_wd_col=ref_wd_col,
        ref_lat=ref_lat,
        ref_long=ref_long,
    )
    post_df = add_waking_scen(
        test_ref_df=post_df,
        test_name=test_name,
        ref_name=ref_name,
        cfg=cfg,
        wf_df=wf_df,
        ref_wd_col=ref_wd_col,
        ref_lat=ref_lat,
        ref_long=ref_long,
    )

    detrend_ws_col = "ref_ws_detrended"
    pre_df = apply_wsratio_v_wd_scen(pre_df, wsratio_v_dir_scen, ref_ws_col=ref_ws_col, ref_wd_col=ref_wd_col)
    plot_apply_wsratio_v_wd_scen(
        pre_df.dropna(subset=[ref_ws_col, test_ws_col, detrend_ws_col, test_pw_col]),
        ref_ws_col=ref_ws_col,
        test_ws_col=test_ws_col,
        detrend_ws_col=detrend_ws_col,
        test_pw_col=test_pw_col,
        test_name=test_name,
        ref_name=ref_name,
        title_end="pre",
        plot_cfg=plot_cfg,
    )
    post_df = apply_wsratio_v_wd_scen(post_df, wsratio_v_dir_scen, ref_ws_col=ref_ws_col, ref_wd_col=ref_wd_col)
    plot_apply_wsratio_v_wd_scen(
        post_df.dropna(subset=[ref_ws_col, test_ws_col, detrend_ws_col, test_pw_col]),
        ref_ws_col=ref_ws_col,
        test_ws_col=test_ws_col,
        detrend_ws_col=detrend_ws_col,
        test_pw_col=test_pw_col,
        test_name=test_name,
        ref_name=ref_name,
        title_end="post",
        plot_cfg=plot_cfg,
    )

    detrend_pre_r2_improvement, detrend_post_r2_improvement = check_applied_detrend(
        test_name=test_name,
        ref_name=ref_name,
        ref_lat=ref_lat,
        ref_long=ref_long,
        pre_df=pre_df,
        post_df=post_df,
        test_ws_col=test_ws_col,
        ref_ws_col=ref_ws_col,
        detrend_ws_col=detrend_ws_col,
        ref_wd_col=ref_wd_col,
        cfg=cfg,
        plot_cfg=plot_cfg,
    )

    plot_pre_post_data_cov(
        cfg=cfg,
        test_name=test_name,
        ref_name=ref_name,
        test_df=test_df,
        test_pw_col=test_pw_col,
        test_ws_col=test_ws_col,
        ref_df=ref_df,
        ref_pw_col=ref_pw_col,
        ref_ws_col=ref_ws_col,
        detrend_ws_col=detrend_ws_col,
        ref_wd_col=ref_wd_col,
        pre_df=pre_df,
        post_df=post_df,
        plot_cfg=plot_cfg,
    )

    if cfg.toggle is not None:
        pre_df, post_df = toggle_pairing_filter(
            pre_df=pre_df,
            post_df=post_df,
            pairing_filter_method=cfg.toggle.pairing_filter_method,
            pairing_filter_timedelta_seconds=cfg.toggle.pairing_filter_timedelta_seconds,
            detrend_ws_col=detrend_ws_col,
            test_pw_col=test_pw_col,
            ref_wd_col=ref_wd_col,
            timebase_s=cfg.timebase_s,
        )

    if plot_cfg is not None:
        plot_yaw_direction_pre_post(
            pre_df=pre_df,
            post_df=post_df,
            test_name=test_name,
            ref_name=ref_name,
            ref_ws_col=ref_ws_col,
            ref_wd_col=ref_wd_col,
            plot_cfg=plot_cfg,
            toggle_name=cfg.toggle.name if cfg.toggle else None,
        )

    pp_results, pp_df = pre_post_pp_analysis_with_reversal_and_bootstrapping(
        cfg=cfg,
        test_wtg=test_wtg,
        ref_name=ref_name,
        lt_df=long_term_df,
        pre_df=pre_df,
        post_df=post_df,
        ws_col=detrend_ws_col,
        pw_col=test_pw_col,
        wd_col=ref_wd_col,
        plot_cfg=plot_cfg,
        test_df=test_df,
        random_seed=random_seed,
    )

    other_results = ref_info | {
        "ref_ws_col": ref_ws_col,
        "distance_m": distance_m,
        "bearing_deg": bearing_deg,
        "ref_max_northing_error_v_reanalysis": ref_max_northing_error_v_reanalysis,
        "ref_max_northing_error_v_wf": ref_max_northing_error_v_wf,
        "ref_max_ws_drift": ref_max_ws_drift,
        "ref_max_ws_drift_pp_period": ref_max_ws_drift_pp_period,
        "ref_powercurve_shift": ref_ops_curve_shift_dict["powercurve_shift"],
        "ref_rpm_shift": ref_ops_curve_shift_dict["rpm_shift"],
        "ref_pitch_shift": ref_ops_curve_shift_dict["pitch_shift"],
        "detrend_pre_r2_improvement": detrend_pre_r2_improvement,
        "detrend_post_r2_improvement": detrend_post_r2_improvement,
        "mean_power_pre": pre_df.dropna(subset=[detrend_ws_col, test_pw_col, ref_wd_col])[test_pw_col].mean(),
        "mean_power_post": post_df.dropna(subset=[detrend_ws_col, test_pw_col, ref_wd_col])[test_pw_col].mean(),
    }

    other_results = other_results | yaw_error_results(
        pre_df=pre_df, post_df=post_df, required_pp_cols=[detrend_ws_col, test_pw_col, ref_wd_col]
    )
    other_results = other_results | yaw_offset_results(
        pre_df=pre_df,
        post_df=post_df,
        required_pp_cols=[detrend_ws_col, test_pw_col, ref_wd_col],
        ref_wd_col=ref_wd_col,
        test_wd_col="test_YawAngleMean",
    )

    other_results["test_ref_warning_counts"] = len(result_manager.stored_warnings)
    result_manager.stored_warnings = []
    return other_results | pp_results


def run_wind_up_analysis(
    inputs: AssessmentInputs,
    random_seed: int = RANDOM_SEED,
) -> pd.DataFrame:
    preprocess_warning_counts = len(result_manager.stored_warnings)
    result_manager.stored_warnings = []

    wf_df = inputs.wf_df
    pc_per_ttype = inputs.pc_per_ttype
    cfg = inputs.cfg
    plot_cfg = inputs.plot_cfg
    pre_post_splitter = inputs.pre_post_splitter

    ref_name_list = sorted(set(cfg.non_wtg_ref_names + [x.name for x in cfg.ref_wtgs]))
    wtgs_to_test = cfg.test_wtgs.copy()
    if len(cfg.ref_wtgs) > 1:
        wtgs_to_test.extend([x for x in cfg.ref_wtgs if x not in cfg.test_wtgs])

    results_per_test_ref = []
    logger.info(f"test turbines: {[x.name for x in cfg.test_wtgs]}")
    logger.info(f"ref list: {ref_name_list}")
    logger.info(f"turbines to test: {[x.name for x in wtgs_to_test]}")
    for test_wtg_counter, test_wtg in enumerate(wtgs_to_test):
        test_name = test_wtg.name
        test_pw_col = "pw_clipped"
        test_ws_col = "ws_est_from_power_only" if cfg.ignore_turbine_anemometer_data else "ws_est_blend"
        test_df = wf_df.loc[test_wtg.name].copy()

        if cfg.filter_all_test_wtgs_together:
            for other_test_wtg in cfg.test_wtgs:
                if other_test_wtg.name == test_name:
                    continue
                pw_na_before = test_df["ActivePowerMean"].isna().sum()
                other_test_df = wf_df.loc[other_test_wtg.name]
                timestamps_to_filter = other_test_df[
                    other_test_df[test_pw_col].isna() | other_test_df[test_ws_col].isna()
                ].index
                cols_to_filter = list({test_pw_col, test_ws_col, "ActivePowerMean", "WindSpeedMean"})
                test_df.loc[timestamps_to_filter, cols_to_filter] = pd.NA
                pw_na_after = test_df["ActivePowerMean"].isna().sum()
                print_filter_stats(
                    filter_name=f"filter_all_test_wtgs_together {other_test_wtg.name}",
                    na_rows=pw_na_after - pw_na_before,
                    total_rows=len(test_df),
                )

        if cfg.use_lt_distribution:
            lt_wtg_df_raw, lt_wtg_df_filt = calc_turbine_lt_dfs_raw_filt(
                wtg_name=test_name,
                cfg=cfg,
                wtg_df=test_df,
                ws_col=test_ws_col,
                pw_col=test_pw_col,
                plot_cfg=plot_cfg,
            )
        else:
            lt_wtg_df_raw = None
            lt_wtg_df_filt = None

        test_df.columns = ["test_" + x for x in test_df.columns]
        test_pw_col = "test_" + test_pw_col
        test_ws_col = "test_" + test_ws_col

        test_max_ws_drift, test_max_ws_drift_pp_period = check_windspeed_drift(
            wtg_df=test_df,
            wtg_name=test_name,
            ws_col=test_ws_col,
            reanalysis_ws_col="test_" + REANALYSIS_WS_COL,
            cfg=cfg,
            plot_cfg=plot_cfg,
        )

        test_df, pre_df, post_df = pre_post_splitter.split(test_df, test_wtg_name=test_name)

        test_ops_curve_shift_dict = check_for_ops_curve_shift(
            pre_df,
            post_df,
            wtg_name=test_name,
            ws_col=test_ws_col,
            pw_col=test_pw_col,
            rpm_col=f"test_{DataColumns.gen_rpm_mean}",
            pt_col=f"test_{DataColumns.pitch_angle_mean}",
        )

        # compare ops curves of pre_df and post_df
        compare_ops_curves_pre_post(
            pre_df=pre_df,
            post_df=post_df,
            test_name=test_name,
            plot_cfg=plot_cfg,
            is_toggle_test=(cfg.toggle is not None),
        )

        test_results = {
            "wind_up_version": wind_up.__version__,
            "test_wtg": test_name,
            "test_pw_col": test_pw_col,
            "lt_wtg_hours_raw": lt_wtg_df_raw["observed_hours"].sum() if lt_wtg_df_raw is not None else 0,
            "lt_wtg_hours_filt": lt_wtg_df_filt["observed_hours"].sum() if lt_wtg_df_filt is not None else 0,
            "test_max_ws_drift": test_max_ws_drift,
            "test_max_ws_drift_pp_period": test_max_ws_drift_pp_period,
            "test_powercurve_shift": test_ops_curve_shift_dict["powercurve_shift"],
            "test_rpm_shift": test_ops_curve_shift_dict["rpm_shift"],
            "test_pitch_shift": test_ops_curve_shift_dict["pitch_shift"],
            "preprocess_warning_counts": preprocess_warning_counts,
            "test_warning_counts": len(result_manager.stored_warnings),
        }
        result_manager.stored_warnings = []

        scada_pc = pc_per_ttype[test_wtg.turbine_type.turbine_type]
        for ref_name_counter, ref_name in enumerate(ref_name_list):
            loop_counter = test_wtg_counter * len(ref_name_list) + ref_name_counter
            logger.info(f"analysing {test_name} {ref_name}, loop_counter={loop_counter}")
            test_ref_results = calc_test_ref_results(
                test_df=test_df,
                pre_df=pre_df,
                post_df=post_df,
                long_term_df=lt_wtg_df_filt,
                wf_df=wf_df,
                toggle_df=inputs.pre_post_splitter.toggle_df,
                test_wtg=test_wtg,
                ref_name=ref_name,
                test_pw_col=test_pw_col,
                test_ws_col=test_ws_col,
                scada_pc=scada_pc,
                cfg=cfg,
                plot_cfg=plot_cfg,
                random_seed=random_seed,
            )
            test_ref_results = test_ref_results | test_results
            logger.info(test_ref_results)
            results_df = pd.DataFrame(test_ref_results, index=[loop_counter])
            results_per_test_ref.append(results_df)
            pd.concat(results_per_test_ref).to_csv(cfg.out_dir / "results_interim.csv")

            try:
                msg = (
                    f"warning summary: preprocess_warning_counts={results_df['preprocess_warning_counts'].iloc[0]}, "
                    f"test_warning_counts={results_df['test_warning_counts'].iloc[0]}, "
                    f"test_ref_warning_counts={results_df['test_ref_warning_counts'].iloc[0]}"
                )
                logger.info(msg)
            except KeyError:
                pass
            logger.info(f"finished analysing {test_name} {ref_name}\n")
    all_results_per_test_ref = pd.concat(results_per_test_ref)
    first_columns = [
        "wind_up_version",
        "time_calculated",
        "preprocess_warning_counts",
        "test_warning_counts",
        "test_ref_warning_counts",
        "test_wtg",
        "test_pw_col",
        "ref",
        "ref_ws_col",
        "uplift_frc",
        "unc_one_sigma_frc",
        "uplift_p95_frc",
        "uplift_p5_frc",
        "pp_data_coverage",
        "distance_m",
        "bearing_deg",
        "unc_one_sigma_noadj_frc",
        "unc_one_sigma_lowerbound_frc",
        "unc_one_sigma_bootstrap_frc",
    ]
    other_columns = [x for x in results_df.columns if x not in first_columns]
    all_results_per_test_ref = all_results_per_test_ref[
        [col for col in first_columns + other_columns if col in all_results_per_test_ref.columns]
    ]
    all_results_per_test_ref.to_csv(
        cfg.out_dir / f"{cfg.assessment_name}_results_per_test_ref_"
        f"{pd.Timestamp.now('UTC').strftime('%Y%m%d_%H%M%S')}.csv",
    )
    return all_results_per_test_ref



import pandas as pd
from matplotlib import pyplot as plt

from wind_up.models import PlotConfig


def plot_lt_ws(
    lt_df: pd.DataFrame,
    *,
    turbine_or_wf_name: str,
    title_end: str,
    plot_cfg: PlotConfig,
    one_turbine: bool = False,
) -> None:
    plt.figure()
    plt.plot(lt_df["bin_mid"], lt_df["hours_per_year"], marker="s", label="hours per year")
    plt.plot(
        lt_df["bin_mid"],
        lt_df["mwh_per_year_per_turbine"],
        marker="s",
        label=f"MWh per year{one_turbine * ' per turbine'}",
    )
    title_end = f" {title_end}" if len(title_end) > 0 else ""
    plot_title = f"{turbine_or_wf_name} long term distribution{title_end}"
    plt.title(plot_title)
    plt.xlabel("wind speed bin middle [m/s]")
    plt.ylabel("hours per year, MWh per year")
    plt.legend()
    plt.grid()
    plt.tight_layout()
    if plot_cfg.show_plots:
        plt.show()
    if plot_cfg.save_plots:
        if one_turbine:
            (plot_cfg.plots_dir / turbine_or_wf_name).mkdir(parents=True, exist_ok=True)
            plt.savefig(plot_cfg.plots_dir / turbine_or_wf_name / f"{plot_title}.png")
        else:
            plt.savefig(plot_cfg.plots_dir / f"{plot_title}.png")
    plt.close()


def plot_lt_ws_raw_filt(
    *,
    lt_df_raw: pd.DataFrame,
    lt_df_filt: pd.DataFrame,
    turbine_or_wf_name: str,
    plot_cfg: PlotConfig,
    one_turbine: bool = False,
) -> None:
    plt.figure()
    plt.plot(lt_df_raw["bin_mid"], lt_df_raw["hours_per_year"], marker="s", label="before filter")
    plt.plot(lt_df_filt["bin_mid"], lt_df_filt["hours_per_year"], marker="s", label="after filter")
    plot_title = f"{turbine_or_wf_name} long term hours distribution"
    plt.title(plot_title)
    plt.xlabel("wind speed bin middle [m/s]")
    plt.ylabel("hours per year")
    plt.legend()
    plt.grid()
    plt.tight_layout()
    if plot_cfg.show_plots:
        plt.show()
    if plot_cfg.save_plots:
        if one_turbine:
            plt.savefig(plot_cfg.plots_dir / turbine_or_wf_name / f"{plot_title}.png")
        else:
            plt.savefig(plot_cfg.plots_dir / f"{plot_title}.png")
    plt.close()

    plt.figure()
    plt.plot(lt_df_raw["bin_mid"], lt_df_raw["mwh_per_year_per_turbine"], marker="s", label="before filter")
    plt.plot(lt_df_filt["bin_mid"], lt_df_filt["mwh_per_year_per_turbine"], marker="s", label="after filter")
    plot_title = f"{turbine_or_wf_name} long term MWh per year{one_turbine * ' per turbine'}"
    plt.title(plot_title)
    plt.xlabel("wind speed bin middle [m/s]")
    plt.ylabel("MWh per year")
    plt.legend()
    plt.grid()
    plt.tight_layout()
    if plot_cfg.show_plots:
        plt.show()
    if plot_cfg.save_plots:
        if one_turbine:
            plt.savefig(plot_cfg.plots_dir / turbine_or_wf_name / f"{plot_title}.png")
        else:
            plt.savefig(plot_cfg.plots_dir / f"{plot_title}.png")
    plt.close()



import logging

logger = logging.getLogger(__name__)


class ResultManager:
    stored_warnings: list[str]

    def __init__(self) -> None:
        self.stored_warnings = []

    def warning(self, msg: str) -> None:
        logger.warning(msg)
        self.stored_warnings.append(msg)


result_manager = ResultManager()



from pathlib import Path

import pandas as pd
import pytest

from tests.conftest import TEST_DATA_FLD
from wind_up.models import WindUpConfig
from wind_up.reanalysis_data import (
    ReanalysisDataset,
    add_reanalysis_data,
    get_dsid_and_dates_from_filename,
)


def test_get_dsid_and_dates_from_filename() -> None:
    assert get_dsid_and_dates_from_filename("ERA5T_47.50N_-3.25E_100m_1hr_19900101_20231031.parquet") == (
        "ERA5T_47.50N_-3.25E_100m_1hr",
        pd.Timestamp("1990-01-01", tz="UTC"),
        pd.Timestamp("2023-10-31", tz="UTC"),
    )


def test_add_reanalysis_data(test_homer_config: WindUpConfig) -> None:
    cfg = test_homer_config
    cfg.lt_first_dt_utc_start = pd.Timestamp("2023-07-01 00:00:00", tz="UTC")
    cfg.analysis_last_dt_utc_start = pd.Timestamp("2023-07-31 23:50:00", tz="UTC")
    wf_df = pd.read_parquet(Path(__file__).parents[0] / "test_data/Homer Wind Farm_July2023_scada_improved.parquet")
    reanalysis_datasets = [
        ReanalysisDataset(id=fp.stem, data=pd.read_parquet(fp))
        for fp in (TEST_DATA_FLD / "reanalysis" / "Homer Wind Farm").glob("*.parquet")
    ]
    wf_df_after = add_reanalysis_data(wf_df, cfg=cfg, plot_cfg=None, reanalysis_datasets=reanalysis_datasets)
    assert len(wf_df_after.dropna(subset=["reanalysis_ws", "reanalysis_wd"])) == 8928
    assert wf_df_after["reanalysis_ws"].mean() == pytest.approx(7.095553315412186)
    assert wf_df_after["reanalysis_ws"].min() == pytest.approx(0.19)
    assert wf_df_after["reanalysis_ws"].max() == pytest.approx(16.35)
    assert wf_df_after["reanalysis_wd"].mean() == pytest.approx(217.18956093189962)
    assert wf_df_after["reanalysis_wd"].min() == pytest.approx(4.4)
    assert wf_df_after["reanalysis_wd"].max() == pytest.approx(359.6)



import itertools
import logging
import math

import numpy as np
import pandas as pd
from scipy.stats import norm

from wind_up.models import PlotConfig
from wind_up.plots.combine_results_plots import plot_combine_results
from wind_up.result_manager import result_manager

logger = logging.getLogger(__name__)


def calc_sigma_ref(rdf: pd.DataFrame, ref_list: list[str]) -> float:
    # calculate the weighted absolute average reference uplift as a lower bound on uncertainty
    ref_uplifts = rdf.loc[(rdf["test_wtg"].isin(ref_list)), "p50_uplift"]
    ref_sigmas = rdf.loc[(rdf["test_wtg"].isin(ref_list)), "sigma_test"]
    weights = 1 / (ref_sigmas**2)
    return (abs(ref_uplifts) * weights).sum() / weights.sum() if weights.sum() > 0 else np.nan


def calc_tdf(trdf: pd.DataFrame, ref_list: list[str], weight_col: str = "unc_weight") -> pd.DataFrame:
    tdf = trdf.groupby("test_wtg").agg(
        p50_uplift=pd.NamedAgg(
            column="uplift_frc",
            aggfunc=lambda x: (x * trdf.loc[x.index, weight_col]).sum() / trdf.loc[x.index, weight_col].sum(),
        ),
        sigma_uncorr=pd.NamedAgg(
            column="unc_one_sigma_frc",
            aggfunc=lambda x: np.sqrt(
                ((x * trdf.loc[x.index, weight_col] / trdf.loc[x.index, weight_col].sum()) ** 2).sum(),
            ),
        ),
        sigma_corr=pd.NamedAgg(
            column="unc_one_sigma_frc",
            aggfunc=lambda x: (x * trdf.loc[x.index, weight_col]).sum() / trdf.loc[x.index, weight_col].sum(),
        ),
        ref_count=pd.NamedAgg(column="uplift_frc", aggfunc=len),
        is_ref=pd.NamedAgg(column="test_wtg", aggfunc=lambda x: x.isin(ref_list).any()),
    )
    tdf["sigma_test"] = (tdf["sigma_uncorr"] + tdf["sigma_corr"]) / 2
    tdf = tdf.sort_values(by=["ref_count", "test_wtg"], ascending=[False, True])
    tdf = tdf.reset_index()
    sigma_ref = calc_sigma_ref(tdf, ref_list)
    tdf["sigma"] = tdf["sigma_test"].clip(lower=sigma_ref)
    tdf["p95_uplift"] = tdf["p50_uplift"] + norm.ppf(0.05) * tdf["sigma"]
    tdf["p5_uplift"] = tdf["p50_uplift"] + norm.ppf(0.95) * tdf["sigma"]
    return tdf


def choose_best_refs(trdf: pd.DataFrame, ref_list: list[str], min_refs: int = 3) -> list[str]:
    ref_count = len(ref_list)
    if ref_count < min_refs:
        msg = "ref_list must have at least min_refs elements"
        raise ValueError(msg)

    rdf = calc_tdf(trdf, ref_list)
    best_sigma_ref = calc_sigma_ref(rdf, ref_list)
    best_ref_list = ref_list.copy()
    ref_count -= 1
    while ref_count >= min_refs:
        for c in itertools.combinations(ref_list, ref_count):
            this_ref_list = list(c)
            this_sigma_ref = calc_sigma_ref(rdf, this_ref_list)
            if this_sigma_ref < best_sigma_ref:
                best_sigma_ref = this_sigma_ref
                best_ref_list = this_ref_list.copy()
        ref_count -= 1
    return best_ref_list


def combine_results(
    trdf: pd.DataFrame,
    *,
    auto_choose_refs: bool = True,
    exclude_refs: list[str] | None = None,
    plot_config: PlotConfig | None = None,
) -> pd.DataFrame:
    if exclude_refs is None:
        exclude_refs = []

    msg = "#" * 78 + "\n# combine results per test turbine\n" + "#" * 78
    logger.info(msg)

    trdf = trdf.copy()

    if trdf.groupby(["test_wtg", "ref"]).size().max() > 1:
        msg = "trdf must have no more than one row per test-ref pair"
        raise ValueError(msg)

    # remove reference predictions of themselves
    trdf = trdf.loc[trdf["test_wtg"] != trdf["ref"], :]

    if len(exclude_refs) > 0:
        logger.info(f"excluding refs {exclude_refs}")
        trdf = trdf.loc[~trdf["test_wtg"].isin(exclude_refs), :]
        trdf = trdf.loc[~trdf["ref"].isin(exclude_refs), :]

    if (trdf["unc_one_sigma_frc"] <= 0).any() or trdf["unc_one_sigma_frc"].isna().any():
        msg = "unc_one_sigma_frc must be positive and non-NaN"
        raise ValueError(msg)

    weight_col = "unc_weight"
    trdf[weight_col] = 1 / (trdf["unc_one_sigma_frc"] ** 2)

    ref_list = sorted(trdf["ref"].unique())

    min_refs = 3
    if auto_choose_refs:
        if len(ref_list) >= min_refs:
            best_ref_list = choose_best_refs(trdf, ref_list, min_refs=min_refs)
            refs_to_remove = [x for x in ref_list if x not in best_ref_list]
            trdf = trdf.loc[~trdf["test_wtg"].isin(refs_to_remove), :]
            trdf = trdf.loc[~trdf["ref"].isin(refs_to_remove), :]
            ref_list = sorted(trdf["ref"].unique())
        else:
            result_manager.warning(f"len(ref_list) < {min_refs}, skipping auto_choose_refs")

    logger.info(f"ref_list = {ref_list}")
    tdf = calc_tdf(trdf, ref_list, weight_col)

    # change column order for readability
    cols = list(tdf.columns)
    first_cols = ["test_wtg", "p50_uplift", "p95_uplift", "p5_uplift", "sigma"]
    cols = first_cols + [x for x in cols if x not in first_cols]
    tdf = tdf[cols]

    if plot_config is not None:
        plot_combine_results(trdf=trdf, tdf=tdf, plot_cfg=plot_config)

    return tdf


def calc_net_uplift(results_per_test_df: pd.DataFrame, *, confidence: float) -> tuple[float, float, float]:
    if results_per_test_df.groupby("test_wtg").size().max() > 1:
        msg = "results_per_test_df must have only one row per test turbine"
        raise ValueError(msg)
    net_p50 = (results_per_test_df["uplift_frc"] * results_per_test_df["mean_power_pre"]).sum() / results_per_test_df[
        "mean_power_pre"
    ].sum()
    net_unc = (
        math.sqrt(((results_per_test_df["unc_one_sigma_frc"] * results_per_test_df["mean_power_pre"]) ** 2).sum())
        / results_per_test_df["mean_power_pre"].sum()
    )
    p_low = (1 - confidence) / 2
    net_p_low = net_p50 + norm.ppf(p_low) * net_unc
    p_high = 1 - p_low
    net_p_high = net_p50 + norm.ppf(p_high) * net_unc
    return net_p50, net_p_low, net_p_high



import logging

import numpy as np
import pandas as pd

from wind_up.models import PlotConfig, WindUpConfig
from wind_up.plots.scada_power_curve_plots import plot_pc_per_ttype, plot_removed_data_per_ttype_and_wtg

logger = logging.getLogger(__name__)


def calc_pc_and_rated_ws_one_ttype(
    df: pd.DataFrame,
    x_col: str,
    y_col: str,
    x_bin_width: float,
) -> tuple[pd.DataFrame, float]:
    x_bin_edges = np.arange(0, df[x_col].max() + x_bin_width, x_bin_width)
    pc = df.groupby(by=pd.cut(df[x_col], bins=x_bin_edges, retbins=False), observed=False).agg(
        x_mean=pd.NamedAgg(column=x_col, aggfunc=lambda x: x.mean()),
        y_mean=pd.NamedAgg(column=y_col, aggfunc=lambda x: x.mean()),
    )
    pc["bin_mid"] = [x.mid for x in pc.index]
    # this rated_ws calculation is not very robust if we are lacking data
    rated_ws_threshold = 0.995
    rated_ws = pc["x_mean"][pc["y_mean"] / pc["y_mean"].max() > rated_ws_threshold].iloc[0]
    logger.info(f"estimated rated wind speed = {rated_ws:.1f} m/s")
    below_rated = pc["x_mean"].fillna(0) < rated_ws
    low_pw_threshold = 0.005
    low_power = pc["y_mean"] / pc["y_mean"].max() < low_pw_threshold
    if (below_rated & low_power).any():
        cutin_ws = pc["x_mean"][below_rated & (pc["y_mean"].fillna(0) / pc["y_mean"].max() < low_pw_threshold)].iloc[-1]
    else:
        cutin_ws = pc["x_mean"].dropna().iloc[0]
    logger.info(f"estimated cut-in wind speed = {cutin_ws:.1f} m/s")

    pc[x_col] = pc["x_mean"].fillna(pc["bin_mid"])
    pc[y_col] = pc["y_mean"]
    pc.loc[pc["bin_mid"] < cutin_ws - x_bin_width / 2, y_col] = 0
    pc.loc[pc["bin_mid"] > rated_ws + x_bin_width / 2, y_col] = pc[y_col].max()

    if pc[y_col].diff().min() < 0:
        msg = f"power curve must be monotonically increasing (x_col = {x_col}, y_col = {y_col})"
        raise RuntimeError(msg)

    return pc, rated_ws


def calc_pc_and_rated_ws(
    cfg: WindUpConfig,
    wf_df: pd.DataFrame,
    x_col: str,
    y_col: str,
    x_bin_width: float,
    plot_cfg: PlotConfig | None,
) -> tuple[dict[str, pd.DataFrame], dict[str, float]]:
    pc_per_ttype: dict[str, pd.DataFrame] = {}
    rated_ws_per_ttype: dict[str, float] = {}
    for ttype in cfg.list_unique_turbine_types():
        wtgs = cfg.list_turbine_ids_of_type(ttype)
        df_ttype = wf_df.loc[wtgs]
        ttype_str = ttype.turbine_type
        original_x_bin_width = x_bin_width
        success = False
        while x_bin_width < (original_x_bin_width * 3):
            try:
                pc_per_ttype[ttype_str], rated_ws_per_ttype[ttype_str] = calc_pc_and_rated_ws_one_ttype(
                    df=df_ttype,
                    x_col=x_col,
                    y_col=y_col,
                    x_bin_width=x_bin_width,
                )
                success = True
                break
            except RuntimeError:
                x_bin_width *= 1.1
                logger.info(f"power curve calculation failed, trying again with larger bin width {x_bin_width:.2f}")
        if not success:
            msg = "power curve calculation failed for all bin widths"
            raise RuntimeError(msg)
    if plot_cfg is not None:
        plot_pc_per_ttype(cfg=cfg, pc_per_ttype=pc_per_ttype, plot_cfg=plot_cfg)
        plot_removed_data_per_ttype_and_wtg(cfg=cfg, wf_df=wf_df, pc_per_ttype=pc_per_ttype, plot_cfg=plot_cfg)
    return pc_per_ttype, rated_ws_per_ttype



import time
from pathlib import Path

import numpy as np
import pandas as pd
import pytest
from pandas.testing import assert_frame_equal

from wind_up.constants import RAW_DOWNTIME_S_COL, RAW_POWER_COL, RAW_WINDSPEED_COL, TIMESTAMP_COL
from wind_up.models import WindUpConfig
from wind_up.scada_funcs import scada_multi_index
from wind_up.waking_state import (
    add_waking_scen,
    add_waking_state,
    calc_bearing,
    calc_distance,
    calc_scen_name_from_wtgs_not_waking,
    get_distance_and_bearing,
    get_iec_upwind_turbines,
    list_wtgs_offline_in_scen,
)


def test_get_bearing(test_homer_config: WindUpConfig) -> None:
    assert calc_bearing(lat1=0, long1=0, lat2=0, long2=1) == 90
    assert calc_bearing(lat1=0, long1=0, lat2=1, long2=0) == 0
    assert calc_bearing(lat1=0, long1=0, lat2=0, long2=-1) == 270
    cfg = test_homer_config
    t1lat = cfg.asset.wtgs[0].latitude
    t1long = cfg.asset.wtgs[0].longitude
    t2lat = cfg.asset.wtgs[1].latitude
    t2long = cfg.asset.wtgs[1].longitude
    expected = 245.02500888680734 - 180
    assert calc_bearing(lat1=t1lat, long1=t1long, lat2=t2lat, long2=t2long) == pytest.approx(expected)
    assert calc_bearing(lat1=t2lat, long1=t2long, lat2=t1lat, long2=t1long) == pytest.approx(
        (expected + 180) % 360,
        rel=1e-4,
    )
    assert calc_bearing(lat1=-t1lat, long1=t1long, lat2=-t2lat, long2=t2long) == pytest.approx((-expected + 180) % 360)
    assert calc_bearing(lat1=t1lat, long1=-t1long, lat2=t2lat, long2=-t2long) == pytest.approx((-expected) % 360)


def test_get_distance(test_homer_config: WindUpConfig) -> None:
    # circumference of earth at equator is about 40075km
    assert calc_distance(lat1=0, long1=0, lat2=0, long2=1 / 100) == pytest.approx(40075 * 1000 / 360 / 100)
    assert calc_distance(lat1=0, long1=0, lat2=0, long2=-1 / 100) == pytest.approx(40075 * 1000 / 360 / 100)
    assert calc_distance(lat1=0, long1=90, lat2=0, long2=90 + 1 / 100) == pytest.approx(40075 * 1000 / 360 / 100)
    cfg = test_homer_config
    t1lat = cfg.asset.wtgs[0].latitude
    t1long = cfg.asset.wtgs[0].longitude
    t2lat = cfg.asset.wtgs[1].latitude
    t2long = cfg.asset.wtgs[1].longitude
    expected = 270.894287973147
    assert calc_distance(lat1=t1lat, long1=t1long, lat2=t2lat, long2=t2long) == pytest.approx(expected)
    assert calc_distance(lat1=t2lat, long1=t2long, lat2=t1lat, long2=t1long) == pytest.approx(expected)
    assert calc_distance(lat1=-t1lat, long1=t1long, lat2=-t2lat, long2=t2long) == pytest.approx(expected)
    assert calc_distance(lat1=t1lat, long1=-t1long, lat2=t2lat, long2=-t2long) == pytest.approx(expected)


def test_get_distance_and_bearing(test_homer_config: WindUpConfig) -> None:
    cfg = test_homer_config
    t1lat = cfg.asset.wtgs[0].latitude
    t1long = cfg.asset.wtgs[0].longitude
    t2lat = cfg.asset.wtgs[1].latitude
    t2long = cfg.asset.wtgs[1].longitude
    expected_bearing = 245.02500888680734 - 180
    expected_distance = 270.894287973147
    start_time = time.perf_counter()
    distance_m, bearing_deg = get_distance_and_bearing(lat1=t1lat, long1=t1long, lat2=t2lat, long2=t2long)
    end_time = time.perf_counter()
    run_time_pre_cache = end_time - start_time
    assert distance_m == pytest.approx(expected_distance)
    assert bearing_deg == pytest.approx(expected_bearing)
    start_time = time.perf_counter()
    reps = 100
    for _x in range(reps):
        distance_m, bearing_deg = get_distance_and_bearing(lat1=t1lat, long1=t1long, lat2=t2lat, long2=t2long)
        assert distance_m == pytest.approx(expected_distance)
        assert bearing_deg == pytest.approx(expected_bearing)
    end_time = time.perf_counter()
    run_time_with_cache = (end_time - start_time) / reps
    assert run_time_with_cache < run_time_pre_cache * 3 / 10


def test_calc_scen_name_from_wtgs_not_waking() -> None:
    assert calc_scen_name_from_wtgs_not_waking(["T1", "T2"]) == "T1 T2 offline"


def test_list_wtgs_offline_in_scen() -> None:
    assert list_wtgs_offline_in_scen("T1 T2 offline") == ["T1", "T2"]


test_get_iec_upwind_turbines_data = [
    (0, 250 - 180, ["HMR_T00", "HMR_T02"]),
    (1, 250 - 180, ["HMR_T00"]),
    (0, 250, []),
    (1, 250, ["HMR_T01"]),
    (0, 250 - 180 - 90, []),
    (1, 250 - 180 - 90, []),
]


@pytest.mark.parametrize(("wtg_idx", "wind_direction", "expected"), test_get_iec_upwind_turbines_data)
def test_get_iec_upwind_turbines(
    wtg_idx: int,
    wind_direction: float,
    expected: list[str],
    test_homer_with_t00_config: WindUpConfig,
) -> None:
    cfg = test_homer_with_t00_config

    upwind_wtgs = get_iec_upwind_turbines(
        latlongs=[(cfg.asset.wtgs[wtg_idx].latitude, cfg.asset.wtgs[wtg_idx].longitude)],
        wind_direction=wind_direction,
        cfg=cfg,
        object_name=cfg.asset.wtgs[wtg_idx].name,
    )
    assert upwind_wtgs == expected


def test_add_waking_scen(test_homer_with_t00_config: WindUpConfig) -> None:
    cfg = test_homer_with_t00_config
    test_name = "HMR_T01"
    ref_name = "HMR_T02"
    ref_wtg = next(x for x in cfg.asset.wtgs if x.name == ref_name)
    ref_lat = ref_wtg.latitude
    ref_long = ref_wtg.longitude

    wtgs = ["HMR_T00", "HMR_T01", "HMR_T02"]
    tstamps = pd.date_range(start="2021-01-01", tz="UTC", periods=3, freq="10min")
    idx = pd.MultiIndex.from_product([wtgs, tstamps], names=["TurbineName", TIMESTAMP_COL])
    wf_df = pd.DataFrame(
        data={
            "YawAngleMean": [249.9 - 180] * len(idx),
            "waking": [True, False, False] + [True] * (len(idx) - 3),
            "not_waking": [False, True, False] + [False] * (len(idx) - 3),
            "unknown_waking": [False, False, True] + [False] * (len(idx) - 3),
        },
        index=idx,
    )
    test_df = wf_df.loc[test_name]
    test_df.columns = ["test_" + x for x in test_df.columns]
    ref_df = wf_df.loc[ref_name]
    ref_df.columns = ["ref_" + x for x in ref_df.columns]
    test_ref_df = test_df.merge(ref_df, how="left", left_index=True, right_index=True)

    expected_df = test_ref_df.copy()
    expected_df["waking_scenario"] = ["none offline", "HMR_T00 offline", "unknown"]

    actual_df = add_waking_scen(
        test_name=test_name,
        ref_name=ref_name,
        test_ref_df=test_ref_df,
        cfg=cfg,
        wf_df=wf_df,
        ref_wd_col="ref_YawAngleMean",
        ref_lat=ref_lat,
        ref_long=ref_long,
    )

    assert_frame_equal(actual_df, expected_df)


def test_add_waking_state_simple_case(test_homer_with_t00_config: WindUpConfig) -> None:
    cfg = test_homer_with_t00_config

    wtgs = ["HMR_T00", "HMR_T01", "HMR_T02"]
    tstamps = pd.date_range(start="2021-01-01", tz="UTC", periods=3, freq="10min")
    idx = pd.MultiIndex.from_product([wtgs, tstamps], names=["TurbineName", TIMESTAMP_COL])
    wf_df = pd.DataFrame(
        data={
            RAW_POWER_COL: [0.009 * cfg.asset.wtgs[0].turbine_type.rated_power_kw, 100, 100] + [500] * (len(idx) - 3),
            RAW_WINDSPEED_COL: [8] * len(idx),
            RAW_DOWNTIME_S_COL: [0] * len(idx),
        },
        index=idx,
    )
    wf_df["WindSpeedMean"] = wf_df[RAW_WINDSPEED_COL]
    wf_df["ActivePowerMean"] = wf_df[RAW_POWER_COL]
    wf_df.iloc[:3, -2:] = pd.NA
    wf_df.iloc[1, 2] = 600 * 0.71  # just enough ShutdownDuration for not_waking
    expected_df = wf_df.copy()
    expected_df["waking"] = [False] * 3 + [True] * (len(idx) - 3)
    expected_df["not_waking"] = [True] * 2 + [False] * (len(idx) - 2)
    expected_df["unknown_waking"] = [False, False, True] + [False] * (len(idx) - 3)
    actual_df = add_waking_state(cfg=cfg, wf_df=wf_df, plot_cfg=None)

    assert_frame_equal(actual_df, expected_df)


def test_add_waking_state_nan_cases(test_homer_with_t00_config: WindUpConfig) -> None:
    cfg = test_homer_with_t00_config

    wtgs = ["HMR_T00", "HMR_T01", "HMR_T02"]
    tstamps = pd.date_range(start="2021-01-01", tz="UTC", periods=3, freq="10min")
    idx = pd.MultiIndex.from_product([wtgs, tstamps], names=["TurbineName", TIMESTAMP_COL])
    pw_value = 100.1
    ws_value = 8.1
    wf_df = pd.DataFrame(
        data={
            RAW_POWER_COL: [np.nan] * 4 + [pw_value] * 4 + [np.nan],
            RAW_WINDSPEED_COL: list([np.nan] * 2 + [ws_value] * 2) * 2 + [np.nan],
            RAW_DOWNTIME_S_COL: [np.nan, 0] * 4 + [np.nan],
        },
        index=idx,
    )
    wf_df["WindSpeedMean"] = wf_df[RAW_WINDSPEED_COL]
    wf_df["ActivePowerMean"] = wf_df[RAW_POWER_COL]
    expected_df = wf_df.copy()
    expected_df["waking"] = [False] * 4 + [True] * 4 + [False]
    expected_df["not_waking"] = [False] * len(idx)
    expected_df["unknown_waking"] = [True] * 4 + [False] * 4 + [True]
    actual_df = add_waking_state(cfg=cfg, wf_df=wf_df, plot_cfg=None)

    assert_frame_equal(actual_df, expected_df)


def test_add_waking_state(test_lsa_t13_config: WindUpConfig) -> None:
    cfg = test_lsa_t13_config
    test_df = pd.read_parquet(Path(__file__).parents[0] / "test_data/LSA_T13_test_df.parquet")
    test_df.columns = test_df.columns.str.replace("test_", "")
    test_df["TurbineName"] = "LSA_T01"
    wf_df = scada_multi_index(test_df)
    expected_df = wf_df.copy()
    wf_df = wf_df.drop(columns=["waking", "not_waking", "unknown_waking"])
    # remove all turbines from cfg apart from LSA_T01
    cfg.asset.wtgs = [cfg.asset.wtgs[0]]
    actual_df = add_waking_state(cfg=cfg, wf_df=wf_df, plot_cfg=None)

    # results not exactly the same because Cp is calculated from the whole dataset and used for waking calc
    # in the original calculation data from the entire wind farm for many years was used
    assert actual_df["not_waking"].sum() - expected_df["not_waking"].sum() == 0
    assert actual_df["waking"].sum() - expected_df["waking"].sum() == 30
    assert actual_df["unknown_waking"].sum() - expected_df["unknown_waking"].sum() == -30



from pathlib import Path

import pandas as pd
from pandas.testing import assert_frame_equal

from wind_up.long_term import calc_turbine_lt_dfs_raw_filt
from wind_up.models import WindUpConfig


def test_calc_turbine_lt_dfs_raw_filt(test_lsa_t13_config: WindUpConfig) -> None:
    cfg = test_lsa_t13_config
    test_df = pd.read_parquet(Path(__file__).parents[0] / "test_data/LSA_T13_test_df.parquet")
    test_df.columns = test_df.columns.str.replace("test_", "")
    test_name = "LSA_T13"
    lt_wtg_df_raw, lt_wtg_df_filt = calc_turbine_lt_dfs_raw_filt(
        wtg_name=test_name,
        cfg=cfg,
        wtg_df=test_df,
        ws_col="WindSpeedMean",
        pw_col="ActivePowerMean",
        plot_cfg=None,
    )

    expected_raw_df = pd.read_parquet(Path(__file__).parents[0] / "test_data/LSA_T13_lt_wtg_df_raw.parquet")
    expected_filt_df = pd.read_parquet(Path(__file__).parents[0] / "test_data/LSA_T13_lt_wtg_df_filt.parquet")
    assert_frame_equal(lt_wtg_df_raw, expected_raw_df)
    assert_frame_equal(lt_wtg_df_filt, expected_filt_df)



import pandas as pd
from matplotlib import pyplot as plt
from scipy.stats import norm

from wind_up.models import PlotConfig


def plot_combine_results(trdf: pd.DataFrame, tdf: pd.DataFrame, plot_cfg: PlotConfig, confidence: float = 0.9) -> None:
    show_plots = plot_cfg.show_plots
    save_plots = plot_cfg.save_plots
    z_score = norm.ppf((1 + confidence) / 2)

    plt.figure(figsize=(10, 6))
    labels = [
        x + "-" + y
        for x, y in zip(
            [x[-1] for x in trdf["test_wtg"].str.split("_")],
            [x[-1] for x in trdf["ref"].str.split("_")],
            strict=True,
        )
    ]
    values = trdf["uplift_frc"] * 100
    yerrs = trdf["unc_one_sigma_frc"] * 100 * z_score
    plt.bar(labels, values, yerr=yerrs, capsize=3)
    plt.xlabel("turbine")
    plt.ylabel("uplift [%]")
    plot_title = f"uplift and {confidence*100:.0f}% CI by test-ref pair"
    plt.title("uplift by test-ref pair")
    plt.grid(axis="y")
    plt.xticks(rotation=90, ha="right")
    plt.tight_layout()
    if show_plots:
        plt.show()
    if save_plots:
        plt.savefig(plot_cfg.plots_dir / f"{plot_title}.png")

    plt.figure()
    labels = tdf["test_wtg"]
    values = tdf["p50_uplift"] * 100
    yerrs = tdf["sigma"] * 100 * z_score
    plt.bar(labels, values, yerr=yerrs, capsize=3)
    plt.xlabel("turbine")
    plt.ylabel("uplift [%]")
    plot_title = f"combined uplift and {confidence*100:.0f}% CI"
    plt.title(plot_title)
    plt.grid(axis="y")
    plt.xticks(rotation=90, ha="right")
    plt.tight_layout()
    if show_plots:
        plt.show()
    if save_plots:
        plt.savefig(plot_cfg.plots_dir / f"{plot_title}.png")



import math

import numpy as np
import pandas as pd
from pandas.testing import assert_frame_equal

from wind_up.constants import TIMESTAMP_COL
from wind_up.main_analysis import toggle_pairing_filter


def test_toggle_pairing_filter_method_none() -> None:
    pre_tstamps = pd.date_range(start="2021-01-01 00:00:00", tz="UTC", periods=9, freq="10min")
    post_tstamps = pd.date_range(start=pre_tstamps.max() + pd.Timedelta("10min"), tz="UTC", periods=9, freq="10min")
    detrend_ws_col = "ref_ws_detrended"
    test_pw_col = "test_pw_clipped"
    ref_wd_col = "ref_YawAngleMean"

    pre_df = pd.DataFrame(
        data={
            detrend_ws_col: [5.1, 5.1, 5.1, 0.0, 0.0, 0.0, np.nan, 5.1, 5.1],
            test_pw_col: [5.1, 5.1, np.nan, 0.0, 0.0, 0.0, 5.1, 5.1, 5.1],
            ref_wd_col: [5.1, 5.1, 5.1, 0.0, 0.0, 0.0, 5.1, 5.1, np.nan],
            "some_col_with_nans": [np.nan] * 9,
        },
        index=pre_tstamps,
    )

    post_df = pd.DataFrame(
        data={
            detrend_ws_col: [5.1, 5.1, 5.1, 0.0, 0.0, 0.0, np.nan, 5.1, 5.1],
            test_pw_col: [5.1, 5.1, np.nan, 0.0, 0.0, 0.0, 5.1, 5.1, 5.1],
            ref_wd_col: [5.1, 5.1, 5.1, 0.0, 0.0, 0.0, 5.1, 5.1, np.nan],
            "some_col_with_nans": [np.nan] * 9,
        },
        index=post_tstamps,
    )

    filt_pre_df, filt_post_df = toggle_pairing_filter(
        pre_df=pre_df,
        post_df=post_df,
        pairing_filter_method="none",
        pairing_filter_timedelta_seconds=0,
        detrend_ws_col=detrend_ws_col,
        test_pw_col=test_pw_col,
        ref_wd_col=ref_wd_col,
        timebase_s=600,
    )
    assert_frame_equal(filt_pre_df, pre_df)
    assert_frame_equal(filt_post_df, post_df)


def test_toggle_pairing_filter_method_any_within_timedelta() -> None:
    pre_tstamps = pd.date_range(start="2021-01-01 00:00:00", tz="UTC", periods=9, freq="10min")
    post_tstamps = pd.date_range(start=pre_tstamps.max() + pd.Timedelta("10min"), tz="UTC", periods=9, freq="10min")
    detrend_ws_col = "ref_ws_detrended"
    test_pw_col = "test_pw_clipped"
    ref_wd_col = "ref_YawAngleMean"

    pre_df = pd.DataFrame(
        data={
            detrend_ws_col: [5.1, 5.1, 5.1, 0.0, 0.0, 0.0, np.nan, 5.1, 5.1],
            test_pw_col: [5.1, 5.1, np.nan, 0.0, 0.0, 0.0, 5.1, 5.1, 5.1],
            ref_wd_col: [5.1, 5.1, 5.1, 0.0, 0.0, 0.0, 5.1, 5.1, np.nan],
            "some_col_with_nans": [np.nan] * 9,
        },
        index=pre_tstamps,
    )
    pre_df.index.name = TIMESTAMP_COL
    post_df = pd.DataFrame(
        data={
            detrend_ws_col: [5.1, 5.1, 5.1, 0.0, 0.0, 0.0, np.nan, 5.1, 5.1],
            test_pw_col: [5.1, 5.1, np.nan, 0.0, 0.0, 0.0, 5.1, 5.1, 5.1],
            ref_wd_col: [5.1, 5.1, 5.1, 0.0, 0.0, 0.0, 5.1, 5.1, np.nan],
            "some_col_with_nans": [np.nan] * 9,
        },
        index=post_tstamps,
    )
    post_df.index.name = TIMESTAMP_COL

    pairing_filter_timedelta_seconds = 50 * 60
    exp_filt_pre_df = pre_df.copy()
    exp_filt_pre_df = exp_filt_pre_df.dropna(subset=[detrend_ws_col, test_pw_col, ref_wd_col])
    exp_filt_post_df = post_df.copy()
    exp_filt_post_df = exp_filt_post_df.dropna(subset=[detrend_ws_col, test_pw_col, ref_wd_col])
    a = exp_filt_pre_df.copy()
    b = exp_filt_post_df.copy()
    # Set the tolerance in minutes (change this value according to your requirements)
    tolerance_minutes = 50

    def copy_of_make_extended_time_index(
        original_index: pd.DatetimeIndex,
        timebase: pd.Timedelta,
        max_timedelta_seconds: int,
    ) -> pd.DatetimeIndex:
        extended_index = original_index
        timedelta_multiple = -math.floor(max_timedelta_seconds / timebase.total_seconds())
        max_timedelta_multiple = math.floor(max_timedelta_seconds / timebase.total_seconds())
        while timedelta_multiple <= max_timedelta_multiple:
            shifted_index = original_index + (timebase * timedelta_multiple)
            extended_index = extended_index.union(shifted_index)
            timedelta_multiple += 1
        return extended_index.sort_values().drop_duplicates()

    exp_filt_pre_df = a[
        [x in copy_of_make_extended_time_index(b.index, pd.Timedelta("10min"), tolerance_minutes * 60) for x in a.index]
    ]
    exp_filt_post_df = b[
        [x in copy_of_make_extended_time_index(a.index, pd.Timedelta("10min"), tolerance_minutes * 60) for x in b.index]
    ]
    filt_pre_df, filt_post_df = toggle_pairing_filter(
        pre_df=pre_df,
        post_df=post_df,
        pairing_filter_method="any_within_timedelta",
        pairing_filter_timedelta_seconds=pairing_filter_timedelta_seconds,
        detrend_ws_col=detrend_ws_col,
        test_pw_col=test_pw_col,
        ref_wd_col=ref_wd_col,
        timebase_s=600,
    )
    assert_frame_equal(filt_pre_df, exp_filt_pre_df)
    assert_frame_equal(filt_post_df, exp_filt_post_df)



from pathlib import Path

import numpy as np
import pandas as pd
from pandas.testing import assert_frame_equal

from wind_up.constants import TIMESTAMP_COL
from wind_up.models import WindUpConfig
from wind_up.scada_funcs import (
    add_pw_clipped,
    filter_bad_pw_ws,
    filter_downtime,
    filter_exclusions,
    filter_missing_rpm_or_pt,
    filter_rpm_and_pt,
    filter_rpm_and_pt_oor_one_ttype,
    filter_stuck_data,
    filter_wrong_yaw,
    filter_yaw_exclusions,
    scada_multi_index,
    wrap_yaw_and_pitch,
)


def test_filter_stuck_data() -> None:
    wtgs = ["MRG_T01", "MRG_T02", "MRG_T03"]
    tstamps = pd.date_range(start="2021-01-01", tz="UTC", periods=3, freq="10min")
    idx = pd.MultiIndex.from_product([wtgs, tstamps], names=["TurbineName", TIMESTAMP_COL])
    adf = pd.DataFrame(
        data={
            "ActivePowerMean": [5.1, 5.1, 5.1, 0.0, 0.0, 0.0, 5.1, 5.1, 5.1],
            "WindSpeedMean": [5.1, 5.1, np.nan, 0.0, 0.0, 0.0, 5.1, 5.1, 5.1],
            "YawAngleMean": [5.1, 5.1, 5.1, 0.0, 0.0, 0.0, 5.1, 5.1, 17],
            "some_col_with_nans": [np.nan] * 9,
        },
        index=idx,
    )
    adf = filter_stuck_data(adf)
    edf = pd.DataFrame(
        data={
            "ActivePowerMean": [5.1, np.nan, np.nan, 0.0, 0.0, 0.0, 5.1, np.nan, 5.1],
            "WindSpeedMean": [5.1, np.nan, np.nan, 0.0, 0.0, 0.0, 5.1, np.nan, 5.1],
            "YawAngleMean": [5.1, np.nan, np.nan, 0.0, 0.0, 0.0, 5.1, np.nan, 17],
            "some_col_with_nans": [np.nan] * 9,
        },
        index=idx,
    )
    assert_frame_equal(edf, adf)


def test_filter_bad_pw_ws() -> None:
    wtgs = ["MRG_T01", "MRG_T02", "MRG_T03"]
    tstamps = pd.date_range(start="2021-01-01", tz="UTC", periods=3, freq="10min")
    idx = pd.MultiIndex.from_product([wtgs, tstamps], names=["TurbineName", TIMESTAMP_COL])
    adf = pd.DataFrame(
        data={
            "ActivePowerMean": [np.nan, np.nan, 0.0, 5.1, 4000, 5.1, -1000, -1001, 4001],
            "WindSpeedMean": [np.nan, 0.1, np.nan, -0.1, 98, 99, 0.0, 0.0, 0.0],
            "some_col_with_nans": [np.nan] * 9,
            "som_col_with_values": [5.1] * 9,
        },
        index=idx,
    )
    adf = filter_bad_pw_ws(adf, max_rated_power=2000)
    edf = pd.DataFrame(
        data={
            "ActivePowerMean": [np.nan, np.nan, np.nan, np.nan, 4000, np.nan, -1000, np.nan, np.nan],
            "WindSpeedMean": [np.nan, np.nan, np.nan, np.nan, 98, np.nan, 0.0, np.nan, np.nan],
            "some_col_with_nans": [np.nan] * 9,
            "som_col_with_values": [np.nan, np.nan, np.nan, np.nan, 5.1, np.nan, 5.1, np.nan, np.nan],
        },
        index=idx,
    )
    assert_frame_equal(edf, adf)


def test_wrap_yaw_and_pitch() -> None:
    wtgs = ["MRG_T01", "MRG_T02", "MRG_T03"]
    tstamps = pd.date_range(start="2021-01-01", tz="UTC", periods=3, freq="10min")
    idx = pd.MultiIndex.from_product([wtgs, tstamps], names=["TurbineName", TIMESTAMP_COL])
    adf = pd.DataFrame(
        data={
            "YawAngleMean": [-181, -180, -179, -1, 0, 1, 359, 360, 361],
            "YawAngleMin": [-182, -181, -180, -2, -1, 0, 358, 359, 360],
            "PitchAngleMean": reversed([-181.1, -180.1, -179.1, -1.1, 0.1, 1.1, 359.1, 360.1, 361.1]),
        },
        index=idx,
    )
    adf = wrap_yaw_and_pitch(adf)
    edf = pd.DataFrame(
        data={
            "YawAngleMean": [179, 180, 181, 359, 0, 1, 359, 0, 1],
            "YawAngleMin": [0, 0, 0, 0, 0, 0, 358, 359, 0],
            "PitchAngleMean": reversed([178.9, 179.9, -179.1, -1.1, 0.1, 1.1, -0.9, 0.1, 1.1]),
        },
        index=idx,
    )
    assert_frame_equal(edf, adf)


def test_filter_wrong_yaw() -> None:
    wtgs = ["MRG_T01", "MRG_T02", "MRG_T03"]
    tstamps = pd.date_range(start="2021-01-01", tz="UTC", periods=3, freq="10min")
    idx = pd.MultiIndex.from_product([wtgs, tstamps], names=["TurbineName", TIMESTAMP_COL])
    adf = pd.DataFrame(
        data={
            "YawAngleMean": [180, 180, 180, 180, 15.1, 344.9, np.nan, 180, 180],
            "YawAngleMin": [170, 170, 190, 1, 1, 1, 180, np.nan, 180],
            "YawAngleMax": [190, 170, 190, 359, 359, 359, 180, 180, np.nan],
        },
        index=idx,
    )
    adf = filter_wrong_yaw(adf)
    edf = pd.DataFrame(
        data={
            "YawAngleMean": [180, 180, 180, np.nan, np.nan, np.nan, np.nan, 180, 180],
            "YawAngleMin": [170, np.nan, np.nan, np.nan, np.nan, np.nan, 180, np.nan, 180],
            "YawAngleMax": [190, np.nan, np.nan, np.nan, np.nan, np.nan, 180, 180, np.nan],
        },
        index=idx,
    )
    assert_frame_equal(edf, adf)


def test_filter_exclusions() -> None:
    wtgs = ["MRG_T01", "MRG_T02", "MRG_T03"]
    tstamps = pd.date_range(start="2021-01-01", tz="UTC", periods=6, freq="10min")
    idx = pd.MultiIndex.from_product([wtgs, tstamps], names=["TurbineName", TIMESTAMP_COL])
    adf = pd.DataFrame(
        data={
            "ActivePowerMean": [1.0] * len(idx),
            "YawAngleMean": [1.0] * len(idx),
        },
        index=idx,
    )
    exclusion_periods_utc = [
        ("MRG_T01", pd.Timestamp("2020-12-31 00:00:00", tz="UTC"), pd.Timestamp("2021-01-01 00:20:00", tz="UTC")),
        ("MRG_T02", pd.Timestamp("2021-01-01 00:29:00", tz="UTC"), pd.Timestamp("2021-01-01 00:31:00", tz="UTC")),
        ("ALL", pd.Timestamp("2021-01-01 00:30:00", tz="UTC"), pd.Timestamp("2022-02-02 00:00:00", tz="UTC")),
    ]
    adf = filter_exclusions(adf.copy(), exclusion_periods_utc)
    expected_values = [
        np.nan,
        np.nan,
        1.0,
        np.nan,
        np.nan,
        np.nan,
        1.0,
        1.0,
        np.nan,
        np.nan,
        np.nan,
        np.nan,
        1.0,
        1.0,
        1.0,
        np.nan,
        np.nan,
        np.nan,
    ]
    edf = pd.DataFrame(
        data={
            "ActivePowerMean": expected_values,
            "YawAngleMean": expected_values,
        },
        index=idx,
    )
    assert_frame_equal(edf, adf)


def test_filter_yaw_exclusions() -> None:
    wtgs = ["MRG_T01", "MRG_T02", "MRG_T03"]
    tstamps = pd.date_range(start="2021-01-01", tz="UTC", periods=6, freq="10min")
    idx = pd.MultiIndex.from_product([wtgs, tstamps], names=["TurbineName", TIMESTAMP_COL])
    adf = pd.DataFrame(
        data={
            "ActivePowerMean": [1.0] * len(idx),
            "YawAngleMean": [1.0] * len(idx),
        },
        index=idx,
    )
    yaw_data_exclusions_utc = [
        ("MRG_T01", pd.Timestamp("2020-12-31 00:00:00", tz="UTC"), pd.Timestamp("2021-01-01 00:20:00", tz="UTC")),
        ("MRG_T02", pd.Timestamp("2021-01-01 00:29:00", tz="UTC"), pd.Timestamp("2021-01-01 00:31:00", tz="UTC")),
        ("ALL", pd.Timestamp("2021-01-01 00:30:00", tz="UTC"), pd.Timestamp("2022-02-02 00:00:00", tz="UTC")),
    ]
    adf = filter_yaw_exclusions(adf.copy(), yaw_data_exclusions_utc)
    edf = pd.DataFrame(
        data={
            "ActivePowerMean": [1.0] * len(idx),
            "YawAngleMean": [
                np.nan,
                np.nan,
                1.0,
                np.nan,
                np.nan,
                np.nan,
                1.0,
                1.0,
                np.nan,
                np.nan,
                np.nan,
                np.nan,
                1.0,
                1.0,
                1.0,
                np.nan,
                np.nan,
                np.nan,
            ],
        },
        index=idx,
    )
    assert_frame_equal(edf, adf)


def test_filter_downtime() -> None:
    wtgs = ["MRG_T01", "MRG_T02"]
    tstamps = pd.date_range(start="2021-01-01", tz="UTC", periods=2, freq="10min")
    idx = pd.MultiIndex.from_product([wtgs, tstamps], names=["TurbineName", TIMESTAMP_COL])
    adf = pd.DataFrame(
        data={
            "ShutdownDuration": [np.nan, -1, 0, 1],
            "ActivePowerMean": [5.1] * 4,
            "some_col_with_nans": [np.nan] * 4,
        },
        index=idx,
    )
    adf = filter_downtime(adf)
    edf = pd.DataFrame(
        data={
            "ShutdownDuration": [np.nan, np.nan, 0, np.nan],
            "ActivePowerMean": [np.nan, np.nan, 5.1, np.nan],
            "some_col_with_nans": [np.nan] * 4,
        },
        index=idx,
    )
    assert_frame_equal(edf, adf)


def test_filter_missing_rpm_or_pt() -> None:
    wtgs = ["MRG_T01", "MRG_T02"]
    tstamps = pd.date_range(start="2021-01-01", tz="UTC", periods=2, freq="10min")
    idx = pd.MultiIndex.from_product([wtgs, tstamps], names=["TurbineName", TIMESTAMP_COL])
    adf = pd.DataFrame(
        data={
            "GenRpmMean": [np.nan, -1, np.nan, -1],
            "PitchAngleMean": [np.nan, np.nan, -1, -1],
            "ActivePowerMean": [5.1] * 4,
            "some_col_with_nans": [np.nan] * 4,
        },
        index=idx,
    )
    adf = filter_missing_rpm_or_pt(adf)
    edf = pd.DataFrame(
        data={
            "GenRpmMean": [np.nan, np.nan, np.nan, -1],
            "PitchAngleMean": [np.nan, np.nan, np.nan, -1],
            "ActivePowerMean": [np.nan, np.nan, np.nan, 5.1],
            "some_col_with_nans": [np.nan] * 4,
        },
        index=idx,
    )
    assert_frame_equal(edf, adf)


def test_filter_rpm_and_pt_oor_one_ttype() -> None:
    wtgs = ["MRG_T01", "MRG_T02"]
    tstamps = pd.date_range(start="2021-01-01", tz="UTC", periods=4, freq="10min")
    idx = pd.MultiIndex.from_product([wtgs, tstamps], names=["TurbineName", TIMESTAMP_COL])
    adf = pd.DataFrame(
        data={
            "GenRpmMean": [799, 800, 1600, 1601] + [1000] * 4,
            "PitchAngleMean": [0] * 4 + [-11, -10, 40, 41],
            "ActivePowerMean": [np.nan, np.nan] + [5.1] * 6,
            "some_col_with_nans": [np.nan] * 8,
        },
        index=idx,
    )
    adf, na_rows = filter_rpm_and_pt_oor_one_ttype(adf, rpm_lower=800, rpm_upper=1600, pt_lower=-10, pt_upper=40)
    edf = pd.DataFrame(
        data={
            "GenRpmMean": [np.nan, 800, 1600, np.nan, np.nan, 1000, 1000, np.nan],
            "PitchAngleMean": [np.nan, 0, 0, np.nan, np.nan, -10, 40, np.nan],
            "ActivePowerMean": [np.nan, np.nan, 5.1, np.nan, np.nan, 5.1, 5.1, np.nan],
            "some_col_with_nans": [np.nan] * 8,
        },
        index=idx,
    )
    assert na_rows == 3
    assert_frame_equal(edf, adf)


def test_add_pw_clipped(test_marge_config: WindUpConfig) -> None:
    wtgs = ["MRG_T01", "MRG_T02", "MRG_T03"]
    tstamps = pd.date_range(start="2021-01-01", tz="UTC", periods=3, freq="10min")
    idx = pd.MultiIndex.from_product([wtgs, tstamps], names=["TurbineName", TIMESTAMP_COL])
    adf = pd.DataFrame(
        data={
            "ActivePowerMean": [-1, 0, 1901, -1, np.nan, 1900, -1, 1900, 1901],
        },
        index=idx,
    )
    cfg = test_marge_config
    cfg.asset.wtgs[2].turbine_type.rated_power_kw = 100
    cfg.asset.wtgs = cfg.asset.wtgs[:3]
    adf = add_pw_clipped(adf, wtgs=cfg.asset.wtgs)
    edf = pd.DataFrame(
        data={
            "ActivePowerMean": [-1, 0, 1901, -1, np.nan, 1900, -1, 1900, 1901],
            "pw_clipped": [0, 0, 1900, 0, np.nan, 1900, 0, 100, 100],
        },
        index=idx,
    )
    assert_frame_equal(edf, adf)


def test_filter_rpm_and_pt(test_marge_config: WindUpConfig) -> None:
    cfg = test_marge_config
    adf = pd.read_parquet(
        Path(__file__).parents[0] / "test_data/smart_data/Marge Wind Farm/Marge Wind Farm_20230101_20230103.parquet"
    )
    adf = scada_multi_index(adf)
    adf = add_pw_clipped(adf, wtgs=cfg.asset.wtgs)
    df_ = filter_rpm_and_pt(
        input_df=adf,
        cfg=cfg,
        plot_cfg=None,
    )

    tt = cfg.asset.wtgs[0].turbine_type
    assert df_["GenRpmMean"].min() >= cfg.get_normal_operation_genrpm_range(tt)[0]
    assert df_["GenRpmMean"].max() <= cfg.get_normal_operation_genrpm_range(tt)[1]
    assert df_["PitchAngleMean"].min() >= cfg.get_normal_operation_pitch_range(tt)[0]
    assert df_["PitchAngleMean"].max() <= cfg.get_normal_operation_pitch_range(tt)[1]

    # the below index locations should definitely be filtered out because of normal operation range
    # MRG_T03	2023-01-02 07:10
    # MRG_T03	2023-01-02 07:20
    # MRG_T03	2023-01-02 07:30
    # MRG_T03	2023-01-02 07:40
    # MRG_T06	2023-01-02 18:00
    # MRG_T06	2023-01-02 18:10
    # MRG_T06	2023-01-02 18:20
    assert df_.loc[("MRG_T03", "2023-01-02 07:10"), :].isna().all()
    assert df_.loc[("MRG_T03", "2023-01-02 07:20"), :].isna().all()
    assert df_.loc[("MRG_T03", "2023-01-02 07:30"), :].isna().all()
    assert df_.loc[("MRG_T03", "2023-01-02 07:40"), :].isna().all()
    assert df_.loc[("MRG_T06", "2023-01-02 18:00"), :].isna().all()
    assert df_.loc[("MRG_T06", "2023-01-02 18:10"), :].isna().all()
    assert df_.loc[("MRG_T06", "2023-01-02 18:20"), :].isna().all()

    # the below are filtered out based on curve filtering so these results could change over time
    assert df_.loc[("MRG_T03", "2023-01-01 12:20"), :].isna().all()
    assert df_.loc[("MRG_T04", "2023-01-02 17:40"), :].isna().all()
    assert df_.loc[("MRG_T06", "2023-01-02 18:30"), :].isna().all()

    # this may be too strict but should catch any major change
    edf = pd.read_parquet(Path(__file__).parents[0] / "test_data/test_filter_rpm_and_pt.parquet")
    assert_frame_equal(edf, df_)



import math

import pandas as pd


def calc_cp(
    power_kw: float | pd.Series,
    ws_ms: float | pd.Series,
    air_density_kgpm3: float | pd.Series,
    rotor_diameter_m: float,
) -> float | pd.Series:
    rotor_area_m2 = math.pi * (rotor_diameter_m / 2) ** 2
    return 1000.0 * power_kw / (0.5 * air_density_kgpm3 * rotor_area_m2 * ws_ms**3)



import logging

import pandas as pd
from matplotlib import pyplot as plt

from wind_up.constants import SCATTER_ALPHA, SCATTER_MARKERSCALE, SCATTER_S
from wind_up.models import PlotConfig
from wind_up.result_manager import result_manager

logger = logging.getLogger(__name__)


def plot_detrend_ws_scatter(
    detrend_df: pd.DataFrame,
    test_name: str,
    ref_name: str,
    test_ws_col: str,
    ref_ws_col: str,
    plot_cfg: PlotConfig,
) -> None:
    plt.figure()
    plt.scatter(detrend_df[ref_ws_col], detrend_df[test_ws_col], s=SCATTER_S, alpha=SCATTER_ALPHA)
    title = f"{test_name} wind speed vs {ref_name} wind speed filtered for detrending"
    plt.title(title)
    plt.xlabel(f"{test_ws_col} [m/s]")
    plt.ylabel(f"{ref_ws_col} [m/s]")
    plt.grid()
    plt.tight_layout()
    if plot_cfg.show_plots:
        plt.show()
    if plot_cfg.save_plots:
        plt.savefig(plot_cfg.plots_dir / test_name / f"{title}.png")
    plt.close()


def plot_detrend_wsratio_v_dir_scen(
    wsratio_v_dir_scen: pd.DataFrame,
    *,
    test_name: str,
    ref_name: str,
    ref_wd_col: str,
    plot_cfg: PlotConfig,
) -> None:
    # figure out top scens to plot
    num_scens_to_plot = 5
    scens_to_plot = (
        wsratio_v_dir_scen.dropna(subset="ws_rom")
        .index.get_level_values("waking_scenario")
        .value_counts()[:num_scens_to_plot]
    )

    plt.figure()
    for scen in scens_to_plot.index:
        scen_df = wsratio_v_dir_scen.loc[scen]
        scen_df = scen_df.reindex(range(360), fill_value=pd.NA)
        plt.plot(scen_df.index, scen_df["hours"], marker=".", label=scen)
    title = f"{test_name} {ref_name} detrending hours"
    plt.title(title)
    plt.legend()
    plt.xlabel(f"{ref_wd_col} [deg]")
    plt.ylabel("hours")
    plt.grid()
    plt.tight_layout()
    if plot_cfg.show_plots:
        plt.show()
    if plot_cfg.save_plots:
        plt.savefig(plot_cfg.plots_dir / test_name / f"{title}.png")
    plt.close()

    plt.figure()
    for scen in scens_to_plot.index:
        scen_df = wsratio_v_dir_scen.loc[scen]
        scen_df = scen_df.reindex(range(360), fill_value=pd.NA)
        plt.plot(scen_df.index, scen_df["ws_rom"], marker=".", label=scen)
    title = f"{test_name} {ref_name} detrending ratio of mean ws"
    plt.title(title)
    plt.legend()
    plt.xlabel(f"{ref_wd_col} [deg]")
    plt.ylabel("ws_rom")
    plt.grid()
    plt.tight_layout()
    if plot_cfg.show_plots:
        plt.show()
    if plot_cfg.save_plots:
        plt.savefig(plot_cfg.plots_dir / test_name / f"{title}.png")
    plt.close()


def plot_apply_wsratio_v_wd_scen(
    p_df: pd.DataFrame,
    ref_ws_col: str,
    test_ws_col: str,
    detrend_ws_col: str,
    test_pw_col: str,
    test_name: str,
    ref_name: str,
    title_end: str,
    plot_cfg: PlotConfig,
) -> None:
    plt.figure()
    r2_before = p_df[ref_ws_col].corr(p_df[test_ws_col]) ** 2
    r2_after = p_df[detrend_ws_col].corr(p_df[test_ws_col]) ** 2
    plt.scatter(
        p_df[ref_ws_col],
        p_df[test_ws_col],
        s=SCATTER_S,
        alpha=SCATTER_ALPHA,
        label=f"before detrend, $r^2$ = {r2_before:.2f}",
    )
    plt.scatter(
        p_df[detrend_ws_col],
        p_df[test_ws_col],
        s=SCATTER_S,
        alpha=SCATTER_ALPHA,
        label=f"after detrend, $r^2$ = {r2_after:.2f}",
    )
    plot_title = f"{test_name} {ref_name} detrended wind speed {title_end}"
    plt.title(plot_title)
    plt.legend(loc="best", markerscale=SCATTER_MARKERSCALE)
    plt.xlabel("ref wind speed [m/s]")
    plt.ylabel(f"{test_ws_col} [m/s]")
    plt.grid()
    plt.tight_layout()
    if plot_cfg.show_plots:
        plt.show()
    if plot_cfg.save_plots:
        plt.savefig(plot_cfg.plots_dir / test_name / f"{plot_title}.png")
    plt.close()

    plt.figure()
    plt.scatter(p_df[ref_ws_col], p_df[test_pw_col], s=SCATTER_S, alpha=SCATTER_ALPHA, label="before detrend")
    plt.scatter(p_df[detrend_ws_col], p_df[test_pw_col], s=SCATTER_S, alpha=SCATTER_ALPHA, label="after detrend")
    plot_title = f"{test_name} {ref_name} detrended power curve {title_end}"
    plt.title(plot_title)
    plt.legend(loc="best", markerscale=SCATTER_MARKERSCALE)
    plt.xlabel("ref wind speed [m/s]")
    plt.ylabel(f"{test_pw_col} [kW]")
    plt.grid()
    plt.tight_layout()
    if plot_cfg.show_plots:
        plt.show()
    if plot_cfg.save_plots:
        plt.savefig(plot_cfg.plots_dir / test_name / f"{plot_title}.png")
    plt.close()


def plot_check_detrend_scatters(
    *,
    pre_df: pd.DataFrame,
    post_df: pd.DataFrame,
    test_name: str,
    ref_name: str,
    test_ws_col: str,
    ref_ws_col: str,
    detrend_ws_col: str,
    plot_cfg: PlotConfig,
) -> None:
    plt.figure()
    plt.subplot(2, 1, 1)
    r2_before = pre_df[ref_ws_col].corr(pre_df[test_ws_col]) ** 2
    r2_after = pre_df[detrend_ws_col].corr(pre_df[test_ws_col]) ** 2
    plt.scatter(
        pre_df[ref_ws_col],
        pre_df[test_ws_col],
        s=SCATTER_S,
        alpha=SCATTER_ALPHA,
        label=f"pre-upgrade before detrend, $r^2$ = {r2_before:.2f}",
    )
    plt.scatter(
        pre_df[detrend_ws_col],
        pre_df[test_ws_col],
        s=SCATTER_S,
        alpha=SCATTER_ALPHA,
        label=f"pre-upgrade after detrend, $r^2$ = {r2_after:.2f}",
    )
    plt.xlabel("ref wind speed before/after detrend [m/s]")
    plt.ylabel(f"{test_ws_col} [m/s]")
    plt.grid()
    plt.legend(fontsize="small", loc="best", markerscale=SCATTER_MARKERSCALE)
    plt.subplot(2, 1, 2)
    r2_before = post_df[ref_ws_col].corr(post_df[test_ws_col]) ** 2
    r2_after = post_df[detrend_ws_col].corr(post_df[test_ws_col]) ** 2
    plt.scatter(
        post_df[ref_ws_col],
        post_df[test_ws_col],
        s=SCATTER_S,
        alpha=SCATTER_ALPHA,
        label=f"post-upgrade before detrend, $r^2$ = {r2_before:.2f}",
    )
    plt.scatter(
        post_df[detrend_ws_col],
        post_df[test_ws_col],
        s=SCATTER_S,
        alpha=SCATTER_ALPHA,
        label=f"post-upgrade after detrend, $r^2$ = {r2_after:.2f}",
    )
    plt.xlabel("ref wind speed before/after detrend [m/s]")
    plt.ylabel(f"{test_ws_col} [m/s]")
    plt.grid()
    plt.legend(fontsize="small", loc="best", markerscale=SCATTER_MARKERSCALE)
    plot_title = f"{test_name} vs {ref_name} wind speed detrend check"
    plt.suptitle(plot_title)
    plt.tight_layout()
    if plot_cfg.show_plots:
        plt.show()
    if plot_cfg.save_plots:
        plt.savefig(plot_cfg.plots_dir / test_name / f"{plot_title}.png")
    plt.close()


def plot_check_wsratio_v_dir(
    pre_wsratio_v_dir: pd.DataFrame,
    post_wsratio_v_dir: pd.DataFrame,
    test_name: str,
    ref_name: str,
    ref_wd_col: str,
    plot_cfg: PlotConfig,
) -> None:
    scen_to_plot = (
        pd.concat([pre_wsratio_v_dir, post_wsratio_v_dir])
        .dropna(subset="ws_rom")
        .groupby("waking_scenario")["hours"]
        .sum()
        .sort_values(ascending=False)
        .index[0]
    )
    try:
        pre_plot_df = pre_wsratio_v_dir.loc[scen_to_plot].reindex(range(360), fill_value=pd.NA)
    except KeyError:
        pre_plot_df = pd.DataFrame(index=range(360), columns=["hours", "ws_rom"])
    try:
        post_plot_df = post_wsratio_v_dir.loc[scen_to_plot].reindex(range(360), fill_value=pd.NA)
    except KeyError:
        post_plot_df = pd.DataFrame(index=range(360), columns=["hours", "ws_rom"])

    plt.figure()
    plt.plot(pre_plot_df.index, pre_plot_df["hours"], marker=".", label=f"{scen_to_plot}, pre-upgrade")
    plt.plot(post_plot_df.index, post_plot_df["hours"], marker=".", label=f"{scen_to_plot}, post-upgrade")
    title = f"{test_name} {ref_name} check detrending hours"
    plt.title(title)
    plt.xlabel(f"{ref_wd_col} [deg]")
    plt.ylabel("hours")
    plt.grid()
    plt.legend()
    plt.tight_layout()
    if plot_cfg.show_plots:
        plt.show()
    if plot_cfg.save_plots:
        plt.savefig(plot_cfg.plots_dir / test_name / f"{title}.png")
    plt.close()

    plt.figure()
    plt.plot(pre_plot_df.index, pre_plot_df["ws_rom"], marker=".", label=f"{scen_to_plot}, pre-upgrade")
    plt.plot(post_plot_df.index, post_plot_df["ws_rom"], marker=".", label=f"{scen_to_plot}, post-upgrade")
    title = f"{test_name} {ref_name} check detrending ratio of mean ws"
    plt.title(title)
    plt.xlabel(f"{ref_wd_col} [deg]")
    plt.ylabel("ws_rom")
    plt.grid()
    plt.legend()
    plt.tight_layout()
    if plot_cfg.show_plots:
        plt.show()
    if plot_cfg.save_plots:
        plt.savefig(plot_cfg.plots_dir / test_name / f"{title}.png")
    plt.close()


def plot_check_applied_detrend(
    *,
    pre_df: pd.DataFrame,
    post_df: pd.DataFrame,
    test_name: str,
    ref_name: str,
    test_ws_col: str,
    ref_ws_col: str,
    detrend_ws_col: str,
    ref_wd_col: str,
    pre_wsratio_v_dir_scen: pd.DataFrame,
    post_wsratio_v_dir_scen: pd.DataFrame,
    plot_cfg: PlotConfig,
) -> None:
    plot_check_detrend_scatters(
        pre_df=pre_df,
        post_df=post_df,
        test_name=test_name,
        ref_name=ref_name,
        test_ws_col=test_ws_col,
        ref_ws_col=ref_ws_col,
        detrend_ws_col=detrend_ws_col,
        plot_cfg=plot_cfg,
    )

    if (
        len(pre_wsratio_v_dir_scen.dropna(subset="ws_rom")) > 0
        or len(post_wsratio_v_dir_scen.dropna(subset="ws_rom")) > 0
    ):
        plot_check_wsratio_v_dir(
            pre_wsratio_v_dir=pre_wsratio_v_dir_scen,
            post_wsratio_v_dir=post_wsratio_v_dir_scen,
            test_name=test_name,
            ref_name=ref_name,
            ref_wd_col=ref_wd_col,
            plot_cfg=plot_cfg,
        )
    else:
        result_manager.warning("check detrending ratio of mean ws plots cannot be made, not enough data")



import matplotlib.pyplot as plt
import pandas as pd

from wind_up.constants import RAW_POWER_COL, RAW_WINDSPEED_COL
from wind_up.models import PlotConfig, WindUpConfig


def plot_detrend_data_cov(
    *,
    cfg: WindUpConfig,
    test_name: str,
    ref_name: str,
    test_df: pd.DataFrame,
    test_ws_col: str,
    ref_df: pd.DataFrame,
    ref_ws_col: str,
    ref_wd_col: str,
    detrend_df: pd.DataFrame,
    plot_cfg: PlotConfig,
) -> None:
    plt.figure(figsize=(12, 8))
    window_hrs = 14 * 24
    rows_per_hour = 3600 / cfg.timebase_s
    expected_rows = rows_per_hour * window_hrs
    xlims = [
        min(cfg.lt_first_dt_utc_start, test_df.index.min()) - pd.Timedelta(f"{window_hrs}h"),
        cfg.analysis_last_dt_utc_start + pd.Timedelta(f"{window_hrs}h"),
    ]
    plt.subplot(3, 1, 1)
    plot_df_raw = (
        test_df.resample(f"{window_hrs}h")["test_" + RAW_POWER_COL, "test_" + RAW_WINDSPEED_COL].count()
        / expected_rows
        * 100
    )
    plot_df = test_df.resample(f"{window_hrs}h")[test_ws_col].count().to_frame(name=test_ws_col) / expected_rows * 100
    plt.plot(plot_df_raw.index, plot_df_raw["test_" + RAW_POWER_COL], label="ActivePowerMean before filter")
    plt.plot(plot_df_raw.index, plot_df_raw["test_" + RAW_WINDSPEED_COL], label="WindSpeedMean before filter")
    plt.plot(plot_df.index, plot_df[test_ws_col], label=test_ws_col)
    plt.xlim(xlims)
    plt.ylim([0, 105])
    plt.ylabel("data coverage [%]")
    plt.grid()
    plt.legend()
    plt.subplot(3, 1, 2)
    plot_df = ref_df.resample(f"{window_hrs}h")[ref_ws_col, ref_wd_col].count() / expected_rows * 100
    plt.plot(plot_df.index, plot_df[ref_ws_col], label=ref_ws_col)
    plt.plot(plot_df.index, plot_df[ref_wd_col], label=ref_wd_col)
    plt.xlim(xlims)
    plt.ylim([0, 105])
    plt.ylabel("data coverage [%]")
    plt.grid()
    plt.legend()
    plt.subplot(3, 1, 3)
    plot_df = (
        detrend_df.resample(f"{window_hrs}h")[test_ws_col].count().to_frame(name=test_ws_col) / expected_rows * 100
    )
    plt.plot(plot_df.index, plot_df[test_ws_col], label=f"detrend {test_ws_col}")
    plt.xlim(xlims)
    plt.ylim([0, 105])
    plt.ylabel("data coverage [%]")
    plt.grid()
    plt.legend()
    plot_title = f"test={test_name} ref={ref_name} detrend data coverage"
    plt.suptitle(plot_title)
    plt.tight_layout()
    if plot_cfg.show_plots:
        plt.show()
    if plot_cfg.save_plots:
        plt.savefig(plot_cfg.plots_dir / test_name / f"{plot_title}.png")
    plt.close()


def plot_pre_post_data_cov(
    *,
    cfg: WindUpConfig,
    test_name: str,
    ref_name: str,
    test_df: pd.DataFrame,
    test_pw_col: str,
    test_ws_col: str,
    ref_df: pd.DataFrame,
    ref_pw_col: str,
    ref_ws_col: str,
    ref_wd_col: str,
    detrend_ws_col: str,
    pre_df: pd.DataFrame,
    post_df: pd.DataFrame,
    plot_cfg: PlotConfig,
) -> None:
    plt.figure(figsize=(12, 8))
    window_hrs = 14 * 24
    rows_per_hour = 3600 / cfg.timebase_s
    expected_rows = rows_per_hour * window_hrs
    xlims = [
        min(cfg.lt_first_dt_utc_start, test_df.index.min()) - pd.Timedelta(f"{window_hrs}h"),
        cfg.analysis_last_dt_utc_start + pd.Timedelta(f"{window_hrs}h"),
    ]
    plt.subplot(3, 1, 1)
    plot_df_raw = (
        test_df.resample(f"{window_hrs}h")["test_" + RAW_POWER_COL, "test_" + RAW_WINDSPEED_COL].count()
        / expected_rows
        * 100
    )
    plot_df = test_df.resample(f"{window_hrs}h")[test_pw_col, test_ws_col].count() / expected_rows * 100
    plt.plot(plot_df_raw.index, plot_df_raw["test_" + RAW_POWER_COL], label="ActivePowerMean before filter")
    plt.plot(plot_df_raw.index, plot_df_raw["test_" + RAW_WINDSPEED_COL], label="WindSpeedMean before filter")
    plt.plot(plot_df.index, plot_df[test_pw_col], label=test_pw_col)
    plt.plot(plot_df.index, plot_df[test_ws_col], label=test_ws_col)
    plt.xlim(xlims)
    plt.ylim([0, 105])
    plt.ylabel("data coverage [%]")
    plt.grid()
    plt.legend()
    plt.subplot(3, 1, 2)
    plot_df = ref_df.resample(f"{window_hrs}h")[ref_pw_col, ref_ws_col, ref_wd_col].count() / expected_rows * 100
    plt.plot(plot_df.index, plot_df[ref_pw_col], label=ref_pw_col)
    plt.plot(plot_df.index, plot_df[ref_ws_col], label=ref_ws_col)
    plt.plot(plot_df.index, plot_df[ref_wd_col], label=ref_wd_col)
    plt.xlim(xlims)
    plt.ylim([0, 105])
    plt.ylabel("data coverage [%]")
    plt.grid()
    plt.legend()
    plt.subplot(3, 1, 3)
    plot_pre_df = pre_df.resample(f"{window_hrs}h")[test_pw_col, detrend_ws_col].count() / expected_rows * 100
    plot_post_df = post_df.resample(f"{window_hrs}h")[test_pw_col, detrend_ws_col].count() / expected_rows * 100
    plt.plot(plot_pre_df.index, plot_pre_df[test_pw_col], label=f"pre {test_pw_col}")
    plt.plot(plot_pre_df.index, plot_pre_df[detrend_ws_col], label=f"pre {detrend_ws_col}")
    plt.plot(plot_post_df.index, plot_post_df[test_pw_col], label=f"post {test_pw_col}")
    plt.plot(plot_post_df.index, plot_post_df[detrend_ws_col], label=f"post {detrend_ws_col}")
    plt.xlim(xlims)
    plt.ylim([0, 105])
    plt.ylabel("data coverage [%]")
    plt.grid()
    plt.legend()
    plot_title = f"test={test_name} ref={ref_name} pre and post data coverage"
    plt.suptitle(plot_title)
    plt.tight_layout()
    if plot_cfg.show_plots:
        plt.show()
    if plot_cfg.save_plots:
        plt.savefig(plot_cfg.plots_dir / test_name / f"{plot_title}.png")
    plt.close()



import os
from pathlib import Path
from typing import IO, Any

import yaml


class Loader(yaml.SafeLoader):
    def __init__(self: "Loader", stream: IO) -> None:
        try:
            self._root = os.path.split(stream.name)[0]
        except AttributeError:
            self._root = os.path.curdir

        super().__init__(stream)

    @property
    def root(self: "Loader") -> str:
        return self._root


def construct_include(loader: Loader, node: yaml.Node) -> Any:  # noqa ANN401
    filename = Path.resolve(Path(loader.root) / Path(loader.construct_scalar(node)))  # type: ignore[arg-type]
    extension = os.path.splitext(filename)[1].lstrip(".")  # noqa PTH122

    with Path.open(filename) as f:
        return yaml.load(f, Loader) if extension in ("yaml", "yml") else "".join(f.readlines())  # noqa S506



import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

from wind_up.models import PlotConfig


def plot_ws_est_gain_xs_one_ttype(
    pc_low_high: pd.DataFrame,
    ttype: str,
    rated_power_kw: float,
    x0: float,
    x1: float,
    x2: float,
    x3: float,
    plot_cfg: PlotConfig,
) -> None:
    plt.figure()
    plt.plot(pc_low_high["x_mean"], pc_low_high["y_low"], label="q low")
    plt.plot(pc_low_high["x_mean"], pc_low_high["y_high"], label="q high")
    plt.vlines(x0, 0, rated_power_kw, label="ws0", color="#C30")
    plt.vlines(x1, 0, rated_power_kw, label="ws1", color="#D20")
    plt.vlines(x2, 0, rated_power_kw, label="ws2", color="#E10")
    plt.vlines(x3, 0, rated_power_kw, label="ws3", color="#F00")
    plot_title = f"{ttype} ws_est_gain xs"
    plt.title(plot_title)
    plt.xlabel("mean wind speed [m/s]")
    plt.ylabel("mean PositiveActivePower [kW]")
    plt.grid()
    plt.legend()
    if plot_cfg.show_plots:
        plt.show()
    if plot_cfg.save_plots:
        plt.savefig(plot_cfg.plots_dir / ttype / f"{plot_title}.png")
    plt.close()


def plot_ws_est_one_ttype_or_wtg(  # noqa C901 PLR0915
    df: pd.DataFrame,
    ttype_or_wtg: str,
    pc_transposed: pd.DataFrame,
    plot_cfg: PlotConfig,
) -> None:
    plt.figure()
    plt.plot(pc_transposed.index, pc_transposed["ws_cp_corrected"])
    plot_title = f"{ttype_or_wtg} transposed power curve"
    plt.title(plot_title)
    plt.xlabel("pw_clipped bin centre [kW]")
    plt.ylabel("ws_cp_corrected [m/s]")
    plt.grid()
    if plot_cfg.show_plots:
        plt.show()
    if plot_cfg.save_plots:
        tdir = plot_cfg.plots_dir / ttype_or_wtg
        tdir.mkdir(exist_ok=True, parents=True)
        plt.savefig(tdir / f"{plot_title}.png")
    plt.close()

    plt.figure()
    plt.scatter(df["ws_cp_corrected"], df["pw_clipped"], s=1, c=df["ws_est_gain"])
    plt.xlabel("ws_cp_corrected [m/s]")
    plt.ylabel("pw_clipped [kW]")
    plot_title = f"{ttype_or_wtg} power curve colored by ws_est_gain"
    plt.title(plot_title)
    plt.colorbar()
    plt.grid()
    if plot_cfg.show_plots:
        plt.show()
    if plot_cfg.save_plots:
        plt.savefig(plot_cfg.plots_dir / ttype_or_wtg / f"{plot_title}.png")
    plt.close()

    plt.figure()
    plt.scatter(df["ws_est_blend"], df["pw_clipped"], s=1)
    plt.xlabel("ws_est_blend [m/s]")
    plt.ylabel("pw_clipped [kW]")
    plot_title = f"{ttype_or_wtg} wind speed estimated from power"
    plt.title(plot_title)
    plt.grid()
    if plot_cfg.show_plots:
        plt.show()
    if plot_cfg.save_plots:
        plt.savefig(plot_cfg.plots_dir / ttype_or_wtg / f"{plot_title}.png")
    plt.close()

    plt.figure()
    plt.scatter(df["WindSpeedMean"], df["ws_est_blend"], s=1, c=df["ws_est_gain"])
    plt.xlabel("WindSpeedMean [m/s]")
    plt.ylabel("ws_est_blend [m/s]")
    plot_title = f"{ttype_or_wtg} wind speed estimated vs original colored by ws_est_gain"
    plt.title(plot_title)
    plt.grid()
    if plot_cfg.show_plots:
        plt.show()
    if plot_cfg.save_plots:
        plt.savefig(plot_cfg.plots_dir / ttype_or_wtg / f"{plot_title}.png")
    plt.close()

    plt.figure()
    try:
        wsbins = np.arange(0, max(df["WindSpeedMean"].max(), df["ws_est_blend"].max()) + 0.5, 0.5)
    except ValueError:
        wsbins = np.arange(0, 25.5, 0.5)
    plt.hist(
        df["WindSpeedMean"],
        bins=wsbins.tolist(),
        label="WindSpeedMean",
    )
    plt.hist(
        df["ws_est_blend"],
        bins=wsbins.tolist(),
        alpha=0.5,
        label="ws_est_blend",
    )
    plt.xlabel("WindSpeedMean, ws_est_blend [m/s]")
    plt.ylabel("count")
    plot_title = f"{ttype_or_wtg} turbines wind speed histograms"
    plt.title(plot_title)
    plt.legend()
    plt.grid()
    if plot_cfg.show_plots:
        plt.show()
    if plot_cfg.save_plots:
        plt.savefig(plot_cfg.plots_dir / ttype_or_wtg / f"{plot_title}.png")
    plt.close()
